from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

import xorq.api as xo
from xorq.backends.pandas.tests.conftest import is_older_than
from xorq.tests.util import assert_frame_equal, assert_series_equal


# SEMI and ANTI are checked in backend tests
mutating_join_type = pytest.mark.parametrize(
    "how",
    ["inner", "left", "right", "outer"],
)


@mutating_join_type
def test_join(how, left, right, df1, df2):
    expr = left.join(right, left.key == right.key, how=how)[
        left, right.other_value, right.key3
    ]
    result = expr.execute()
    expected = pd.merge(df1, df2, how=how, on="key")
    assert_frame_equal(result[expected.columns], expected)


def test_cross_join(left, right, df1, df2):
    expr = left.cross_join(right)[left, right.other_value, right.key3]
    result = expr.execute()
    expected = pd.merge(
        df1.assign(dummy=1), df2.assign(dummy=1), how="inner", on="dummy"
    ).rename(columns={"key_x": "key"})
    del expected["dummy"], expected["key_y"]
    assert_frame_equal(result[expected.columns], expected)


@mutating_join_type
def test_join_project_left_table(how, left, right, df1, df2):
    expr = left.join(right, left.key == right.key, how=how)[left, right.key3]
    result = expr.execute()
    expected = pd.merge(df1, df2, how=how, on="key")[list(left.columns) + ["key3"]]
    assert_frame_equal(result[expected.columns], expected)


def test_cross_join_project_left_table(left, right, df1, df2):
    expr = left.cross_join(right)[left, right.key3]
    result = expr.execute()
    expected = pd.merge(
        df1.assign(dummy=1), df2.assign(dummy=1), how="inner", on="dummy"
    ).rename(columns={"key_x": "key"})[list(left.columns) + ["key3"]]
    assert_frame_equal(result[expected.columns], expected)


@pytest.mark.parametrize(
    "how",
    [
        pytest.param(
            "inner",
            marks=pytest.mark.xfail(
                condition=is_older_than("pandas", "2.0.0"), reason="different indices"
            ),
        ),
        "left",
        "right",
        "outer",
    ],
)
def test_join_with_multiple_predicates(how, left, right, df1, df2):
    expr = left.join(right, [left.key == right.key, left.key2 == right.key3], how=how)[
        left, right.key3, right.other_value
    ]
    result = expr.execute()
    expected = pd.merge(
        df1,
        df2,
        how=how,
        left_on=["key", "key2"],
        right_on=["key", "key3"],
        suffixes=("_left", "_right"),
    ).reset_index(drop=True)

    expected_columns = ["key", "value", "key2", "key3", "other_value"]
    expected = expected[expected_columns]
    if how == "right":
        # the ibis expression references the `key` column from the left table
        # which is not present in the result of the right join, but pandas
        # includes the column from the right table
        expected["key"] = pd.Series([np.nan, np.nan, np.nan], dtype=object)
    elif how == "outer":
        expected["key"] = pd.Series(["a", np.nan, "b", np.nan, "c", "d"], dtype=object)

    assert list(result.columns) == expected_columns
    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "how",
    [
        pytest.param(
            "inner",
            marks=pytest.mark.xfail(
                condition=is_older_than("pandas", "2.0.0"), reason="different indices"
            ),
        ),
        "left",
        "right",
        "outer",
    ],
)
def test_join_with_multiple_predicates_written_as_one(how, left, right, df1, df2):
    predicate = (left.key == right.key) & (left.key2 == right.key3)
    expr = left.join(right, predicate, how=how)[left, right.key3, right.other_value]
    result = expr.execute()
    expected = pd.merge(
        df1, df2, how=how, left_on=["key", "key2"], right_on=["key", "key3"]
    ).reset_index(drop=True)

    if how == "right":
        expected["key"] = pd.Series([np.nan, np.nan], dtype=object)
    elif how == "outer":
        expected["key"] = pd.Series(["a", np.nan, "b", np.nan, "c", "d"], dtype=object)

    assert_frame_equal(result[expected.columns], expected)


@mutating_join_type
def test_join_with_invalid_predicates(how, left, right):
    predicate = (left.key == right.key) & (left.key2 <= right.key3)
    expr = left.join(right, predicate, how=how)
    with pytest.raises(TypeError):
        expr.execute()

    predicate = left.key >= right.key
    expr = left.join(right, predicate, how=how)
    with pytest.raises(TypeError):
        expr.execute()


@mutating_join_type
@pytest.mark.xfail(reason="Hard to detect this case")
def test_join_with_duplicate_non_key_columns(how, left, right):
    left = left.mutate(x=left.value * 2)
    right = right.mutate(x=right.other_value * 3)
    expr = left.join(right, left.key == right.key, how=how)

    # This is undefined behavior because `x` is duplicated. This is difficult
    # to detect
    with pytest.raises(ValueError):
        expr.execute()


@mutating_join_type
def test_join_with_duplicate_non_key_columns_not_selected(how, left, right, df1, df2):
    left = left.mutate(x=left.value * 2)
    right = right.mutate(x=right.other_value * 3)
    right = right[["key", "other_value"]]
    expr = left.join(right, left.key == right.key, how=how)[left, right.other_value]
    result = expr.execute()
    expected = pd.merge(
        df1.assign(x=df1.value * 2),
        df2[["key", "other_value"]],
        how=how,
        on="key",
    )
    assert_frame_equal(result[expected.columns], expected)


@mutating_join_type
def test_join_with_post_expression_selection(how, left, right, df1, df2):
    join = left.join(right, left.key == right.key, how=how)
    expr = join[left.key, left.value, right.other_value]
    result = expr.execute()
    expected = pd.merge(df1, df2, on="key", how=how)[["key", "value", "other_value"]]
    assert_frame_equal(result[expected.columns], expected)


@mutating_join_type
def test_join_with_post_expression_filter(how, left):
    lhs = left[["key", "key2"]]
    rhs = left[["key2", "value"]]

    joined = lhs.join(rhs, "key2", how=how)
    projected = joined[lhs, rhs.value]
    expr = projected[projected.value == 4]
    result = expr.execute()

    df1 = lhs.execute()
    df2 = rhs.execute()
    expected = pd.merge(df1, df2, on="key2", how=how)
    expected = expected.loc[expected.value == 4].reset_index(drop=True)

    assert_frame_equal(result, expected)


@mutating_join_type
def test_multi_join_with_post_expression_filter(how, left, df1):
    lhs = left[["key", "key2"]]
    rhs = left[["key2", "value"]]
    rhs2 = left[["key2", "value"]].rename(value2="value")

    joined = lhs.join(rhs, "key2", how=how)
    projected = joined[lhs, rhs.value]
    filtered = projected[projected.value == 4]

    joined2 = filtered.join(rhs2, "key2")
    projected2 = joined2[filtered.key, rhs2.value2]
    expr = projected2[projected2.value2 == 3]

    result = expr.execute()

    df1 = lhs.execute()
    df2 = rhs.execute()
    df3 = rhs2.execute()
    expected = pd.merge(df1, df2, on="key2", how=how)
    expected = expected.loc[expected.value == 4].reset_index(drop=True)
    expected = pd.merge(expected, df3, on="key2")[["key", "value2"]]
    expected = expected.loc[expected.value2 == 3].reset_index(drop=True)

    assert_frame_equal(result, expected)


@mutating_join_type
def test_join_with_non_trivial_key(how, left, right, df1, df2):
    # also test that the order of operands in the predicate doesn't matter
    join = left.join(right, right.key.length() == left.key.length(), how=how)
    expr = join[left.key, left.value, right.other_value]
    result = expr.execute()

    expected = (
        pd.merge(
            df1.assign(key_len=df1.key.str.len()),
            df2.assign(key_len=df2.key.str.len()),
            on="key_len",
            how=how,
        )
        .drop(["key_len", "key_y", "key2", "key3"], axis=1)
        .rename(columns={"key_x": "key"})
    )
    assert_frame_equal(result[expected.columns], expected)


@mutating_join_type
def test_join_with_non_trivial_key_project_table(how, left, right, df1, df2):
    # also test that the order of operands in the predicate doesn't matter
    join = left.join(right, right.key.length() == left.key.length(), how=how)
    expr = join[left, right.other_value]
    expr = expr[expr.key.length() == 1]
    result = expr.execute()

    expected = (
        pd.merge(
            df1.assign(key_len=df1.key.str.len()),
            df2.assign(key_len=df2.key.str.len()),
            on="key_len",
            how=how,
        )
        .drop(["key_len", "key_y", "key2", "key3"], axis=1)
        .rename(columns={"key_x": "key"})
    )
    expected = expected.loc[expected.key.str.len() == 1]
    assert_frame_equal(result[expected.columns], expected)


@mutating_join_type
def test_join_with_project_right_duplicate_column(client, how, left, df1, df3):
    # also test that the order of operands in the predicate doesn't matter
    right = client.table("df3")
    join = left.join(right, ["key"], how=how)
    expr = join[left.key, right.key2, right.other_value]
    result = expr.execute()

    expected = (
        pd.merge(df1, df3, on="key", how=how)
        .drop(["key2_x", "key3", "value"], axis=1)
        .rename(columns={"key2_y": "key2"})
    )
    assert_frame_equal(result[expected.columns], expected)


def test_join_with_window_function(players_base, players_df, batting, batting_df):
    players = players_base

    # this should be semi_join
    tbl = batting.left_join(players, ["playerID"])
    t = tbl[batting.G, batting.playerID, batting.teamID]
    expr = t.group_by(t.teamID).mutate(
        team_avg=lambda d: d.G.mean(),
        demeaned_by_player=lambda d: d.G - d.G.mean(),
    )
    result = expr.execute()

    expected = pd.merge(
        batting_df, players_df[["playerID"]], on="playerID", how="left"
    )[["G", "playerID", "teamID"]]
    team_avg = expected.groupby("teamID").G.transform("mean")
    expected = expected.assign(
        team_avg=team_avg, demeaned_by_player=lambda df: df.G - team_avg
    )

    assert_frame_equal(result[expected.columns], expected)


merge_asof_minversion = pytest.mark.skipif(
    pd.__version__ < "0.19.2",
    reason="at least pandas-0.19.2 required for merge_asof",
)


@merge_asof_minversion
def test_asof_join(time_left, time_right, time_df1, time_df2):
    expr = time_left.asof_join(time_right, "time")
    result = expr.execute()
    expected = pd.merge_asof(time_df1, time_df2, on="time")
    assert_frame_equal(result[expected.columns], expected)
    with pytest.raises(AssertionError):
        assert_series_equal(result["time"], result["time_right"])


@merge_asof_minversion
def test_asof_join_predicate(time_left, time_right, time_df1, time_df2):
    expr = time_left.asof_join(time_right, time_left.time == time_right.time)
    result = expr.execute()
    expected = pd.merge_asof(
        time_df1, time_df2, on="time", direction="nearest", allow_exact_matches=True
    )
    assert_frame_equal(result[expected.columns], expected)
    with pytest.raises(AssertionError):
        assert_series_equal(result["time"], result["time_right"])


@merge_asof_minversion
def test_keyed_asof_join(
    time_keyed_left, time_keyed_right, time_keyed_df1, time_keyed_df2
):
    expr = time_keyed_left.asof_join(time_keyed_right, "time", predicates="key")
    expr = expr.select(time_keyed_left, time_keyed_right.other_value)
    result = expr.execute()
    expected = pd.merge_asof(time_keyed_df1, time_keyed_df2, on="time", by="key")
    assert_frame_equal(result[expected.columns], expected)


@merge_asof_minversion
def test_keyed_asof_join_with_tolerance(
    time_keyed_left, time_keyed_right, time_keyed_df1, time_keyed_df2
):
    expr = time_keyed_left.asof_join(
        time_keyed_right, "time", predicates="key", tolerance=2 * xo.interval(days=1)
    )
    result = expr.execute()
    expected = pd.merge_asof(
        time_keyed_df1,
        time_keyed_df2,
        on="time",
        by="key",
        tolerance=pd.Timedelta("2D"),
    )
    assert_frame_equal(result[expected.columns], expected)
    with pytest.raises(AssertionError):
        assert_series_equal(result["time"], result["time_right"])
    with pytest.raises(AssertionError):
        assert_series_equal(result["key"], result["key_right"])


@merge_asof_minversion
def test_asof_join_overlapping_non_predicate(
    time_keyed_left, time_keyed_right, time_keyed_df1, time_keyed_df2
):
    # Add a junk column with a colliding name
    time_keyed_left = time_keyed_left.mutate(
        collide=time_keyed_left.key + time_keyed_left.value
    )
    time_keyed_right = time_keyed_right.mutate(
        collide=time_keyed_right.key + time_keyed_right.other_value
    )
    time_keyed_df1.assign(collide=time_keyed_df1["key"] + time_keyed_df1["value"])
    time_keyed_df2.assign(collide=time_keyed_df2["key"] + time_keyed_df2["other_value"])

    expr = time_keyed_left.asof_join(
        time_keyed_right, on=("time", "time"), predicates=[("key", "key")]
    )
    result = expr.execute()
    expected = pd.merge_asof(
        time_keyed_df1, time_keyed_df2, on="time", by="key", suffixes=("", "_right")
    )
    assert_frame_equal(result[expected.columns], expected)


@pytest.mark.parametrize(
    "how",
    [
        "left",
        "right",
        "inner",
        "outer",
    ],
)
@pytest.mark.parametrize(
    "func",
    [
        pytest.param(lambda join: join["a0", "a1"], id="tuple"),
        pytest.param(lambda join: join[["a0", "a1"]], id="list"),
        pytest.param(lambda join: join.select(["a0", "a1"]), id="select"),
    ],
)
def test_select_on_unambiguous_join(how, func):
    df_t = pd.DataFrame({"a0": [1, 2, 3], "b1": list("aab")})
    df_s = pd.DataFrame({"a1": [2, 3, 4], "b2": list("abc")})
    con = xo.pandas.connect({"t": df_t, "s": df_s})
    t = con.table("t")
    s = con.table("s")
    method = getattr(t, f"{how}_join")
    join = method(s, t.b1 == s.b2)
    expected = pd.merge(df_t, df_s, left_on=["b1"], right_on=["b2"], how=how)[
        ["a0", "a1"]
    ]
    assert not expected.empty
    expr = func(join)
    result = expr.execute()
    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "func",
    [
        pytest.param(lambda join: join["a0", "a1"], id="tuple"),
        pytest.param(lambda join: join[["a0", "a1"]], id="list"),
        pytest.param(lambda join: join.select(["a0", "a1"]), id="select"),
    ],
)
@merge_asof_minversion
def test_select_on_unambiguous_asof_join(func):
    df_t = pd.DataFrame({"a0": [1, 2, 3], "b1": pd.date_range("20180101", periods=3)})
    df_s = pd.DataFrame({"a1": [2, 3, 4], "b2": pd.date_range("20171230", periods=3)})
    con = xo.pandas.connect({"t": df_t, "s": df_s})
    t = con.table("t")
    s = con.table("s")
    join = t.asof_join(s, t.b1 == s.b2)
    expected = pd.merge_asof(df_t, df_s, left_on=["b1"], right_on=["b2"])[["a0", "a1"]]
    assert not expected.empty
    expr = func(join)
    result = expr.execute()
    assert_frame_equal(result, expected)


def test_outer_join():
    df = pd.DataFrame({"test": [1, 2, 3], "name": ["a", "b", "c"]})
    df_2 = pd.DataFrame({"test_2": [1, 5, 6], "name_2": ["d", "e", "f"]})

    conn = xo.pandas.connect({"df": df, "df_2": df_2})

    ibis_table_1 = conn.table("df")
    ibis_table_2 = conn.table("df_2")

    joined = ibis_table_1.outer_join(
        ibis_table_2,
        predicates=ibis_table_1["test"] == ibis_table_2["test_2"],
    )
    result = joined.execute()
    expected = pd.merge(
        df,
        df_2,
        left_on="test",
        right_on="test_2",
        how="outer",
    )
    assert_frame_equal(result, expected)


def test_mutate_after_join():
    # GH3090
    df = pd.DataFrame(
        {
            "p_Order_Priority": ["C", "H", "L", "M"],
            "p_count": [9, 9, 15, 11],
            "p_density": [0.204545, 0.204545, 0.340909, 0.250000],
        }
    )
    df_2 = pd.DataFrame(
        {
            "q_Order_Priority": ["C", "H", "L", "M"],
            "q_count": [13, 21, 12, 10],
            "q_density": [0.232143, 0.375000, 0.214286, 0.178571],
        }
    )

    conn = xo.pandas.connect({"df": df, "df_2": df_2})

    ibis_table_1 = conn.table("df")
    ibis_table_2 = conn.table("df_2")

    joined = ibis_table_1.outer_join(
        ibis_table_2,
        predicates=(
            ibis_table_1["p_Order_Priority"] == ibis_table_2["q_Order_Priority"]
        ),
    )

    joined = joined.mutate(
        bins=(
            joined["p_Order_Priority"]
            .isnull()
            .ifelse(joined["q_Order_Priority"], joined["p_Order_Priority"])
        ),
        p_count=joined["p_count"].fill_null(0),
        q_count=joined["q_count"].fill_null(0),
        p_density=joined.p_density.fill_null(1e-10),
        q_density=joined.q_density.fill_null(1e-10),
        features=xo.literal("Order_Priority"),
    )

    expected = pd.DataFrame(
        {
            "p_Order_Priority": list("CHLM"),
            "p_count": [9, 9, 15, 11],
            "p_density": [0.204545, 0.204545, 0.340909, 0.250000],
            "q_Order_Priority": list("CHLM"),
            "q_count": [13, 21, 12, 10],
            "q_density": [0.232143, 0.375000, 0.214286, 0.178571],
            "bins": list("CHLM"),
            "features": ["Order_Priority"] * 4,
        }
    )
    result = joined.execute()
    assert_frame_equal(result, expected)


@pytest.fixture
def tracts_df():
    return pd.DataFrame(
        [[1, 1], [2, 1], [3, 2], [4, 2], [5, 3], [6, 4]],
        columns=["tract_id", "tract_farm_id"],
    )


@pytest.fixture
def fields_df():
    vals = [
        [1, 1, "[(0, 2), (1, 3), (2, 0), (3, 1)]"],
        [2, 1, "[(2, 2), (3, 2), (3, 1)]"],
        [3, 2, "[(0, 1), (-1, 0), (-2, 0), (-2, 1)]"],
        [4, 3, "[(0, 1), (1, 1), (1, 2), (0, 2)]"],
        [5, 3, "[(1, 0), (2, 0), (2, 3), (1, 3)]"],
        [6, 3, "[(2, 0), (3, 0), (3, 2), (2, 2)]"],
        [7, 4, "[(-1, -1), (0, -1), (0, -2)]"],
        [8, 4, "[(1, 0), (1, -2), (0, -2), (0, -1)]"],
        [
            9,
            5,
            str(
                [
                    (1, 0),
                    (2, 0),
                    (1, -1),
                    (1, -2),
                    (-1, -2),
                    (-1, -1),
                    (-2, 0),
                    (-1, 0),
                    (0, -1),
                ]
            ),
        ],
        [10, 6, "[(-1, 2), (0, 2), (0, 0), (-1, 0)]"],
        [11, 6, "[(0, 2), (1, 2), (1, 1), (0, 1)]"],
    ]
    return pd.DataFrame(
        vals,
        columns=["field_id", "field_tract_id", "field_vertices"],
    )


@pytest.fixture
def harvest_df():
    vals = [
        [1, 1, 1, 1, 1, 65.80],
        [2, 2, 1, 2, 2, 5750.00],
        [3, 3, 1, 1, 1, 59.85],
        [4, 4, 2, 2, 2, 10100.00],
        [5, 5, 2, 1, 1, 90.30],
        [6, 6, 2, 2, 2, 21000.00],
        [7, 7, 2, 2, 2, 5150.00],
        [8, 8, 2, 1, 1, 53.55],
        [9, 9, 3, 1, 1, 147.00],
        [10, 10, 4, 1, 1, 70.70],
        [11, 11, 4, 2, 2, 9600.00],
        [12, 1, 1, 2, 4, 22800.00],
        [13, 2, 1, 1, 3, 19.25],
        [14, 3, 1, 2, 4, 13050.00],
        [15, 4, 2, 1, 3, 31.15],
        [16, 5, 2, 2, 4, 33000.00],
        [17, 6, 2, 1, 3, 64.40],
        [18, 7, 2, 1, 3, 16.45],
        [19, 8, 2, 2, 4, 15000.00],
        [20, 9, 3, 2, 4, 38400.00],
        [21, 10, 4, 2, 4, 19800.00],
        [22, 11, 4, 1, 3, 34.30],
    ]

    return pd.DataFrame(
        vals,
        columns=[
            "harvest_id",
            "harvest_field_id",
            "harvest_farmer_group_id",
            "harvest_crop_id",
            "harvest_date_id",
            "harvest_value",
        ],
    )


def test_multijoin(tracts_df, fields_df, harvest_df):
    conn = xo.pandas.connect(
        {
            "tracts": tracts_df,
            "fields": fields_df,
            "harvest": harvest_df,
        }
    )

    tracts, fields, harvest = map(conn.table, "tracts fields harvest".split())

    fielded = harvest.inner_join(
        fields,
        harvest.harvest_field_id == fields.field_id,
    )
    tracted = fielded.inner_join(
        tracts,
        fielded.field_tract_id == tracts.tract_id,
    )
    result = tracted.execute()

    fielded_df = pd.merge(
        harvest_df,
        fields_df,
        left_on="harvest_field_id",
        right_on="field_id",
    )
    expected = pd.merge(
        fielded_df,
        tracts_df,
        left_on="field_tract_id",
        right_on="tract_id",
    )

    assert_frame_equal(result, expected)


def test_chain_join():
    test_df1 = pd.DataFrame({"id": ["1", "1"], "value": ["a", "a"]})
    test_df2 = pd.DataFrame({"id": ["1", "1"], "value": ["z", "z"]})
    test_df3 = pd.DataFrame({"id": ["1", "1"], "value": ["z1", "z1"]})

    conn = xo.pandas.connect({"df1": test_df1, "df2": test_df2, "df3": test_df3})

    t1 = conn.table("df1")
    t2 = conn.table("df2")
    t3 = conn.table("df3")

    expr = (
        t1.join(t2, t1.id == t2.id)
        .join(t3, t1.id == t3.id)
        .select(t1.id, t1.value, t2.value.name("value2"), t3.value.name("value3"))
    )
    result = expr.execute()

    n = len(test_df1) * len(test_df2) * len(test_df3)
    expected = pd.DataFrame(
        {
            "id": ["1"] * n,
            "value": ["a"] * n,
            "value2": ["z"] * n,
            "value3": ["z1"] * n,
        }
    )
    assert_frame_equal(result, expected)



from __future__ import annotations

from public import public

from xorq.vendor.ibis.expr.types.generic import Column, Scalar, Value


@public
class UUIDValue(Value):
    pass


@public
class UUIDScalar(Scalar, UUIDValue):
    pass


@public
class UUIDColumn(Column, UUIDValue):
    pass



import operator

import pytest
import toolz

import xorq.api as xo
from xorq.caching import (
    ParquetCache,
    ParquetSnapshotCache,
)
from xorq.common.utils.defer_utils import deferred_read_parquet
from xorq.expr.relations import into_backend


rate_to_rate_str = toolz.compose(operator.methodcaller("replace", ".", "p"), str)


def asof_join_flight_data(con, tail, flight, airborne_only=True):
    """Create an expression for a particular flight's data"""

    def rate_to_parquet(tail, flight, rate):
        base_url = (
            f"https://nasa-avionics-data-ml.s3.us-east-2.amazonaws.com/{tail}_parquet"
        )
        filename = f"{flight}.{rate_to_rate_str(rate)}.parquet"
        return f"{base_url}/{filename}"

    rates = (0.25, 1.0, 2.0, 4.0, 8.0, 16.0)
    ts = (
        deferred_read_parquet(parquet_path, con, rate_to_rate_str(rate)).mutate(
            flight=xo.literal(flight)
        )
        for rate, parquet_path in (
            (rate, rate_to_parquet(tail, flight, rate))
            for rate in sorted(rates, reverse=True)
        )
    )
    db_con = xo.duckdb.connect()
    (expr, *others) = (
        into_backend(t, db_con, name=f"flight-{flight}-{t.op().parent.name}")
        for t in ts
    )
    for other in others:
        expr = expr.asof_join(other, on="time").drop(["time_right", "flight_right"])
    if airborne_only:
        expr = expr[lambda t: t.GS != 0]
    return expr


@pytest.mark.skip
@pytest.mark.parametrize("cls", [ParquetSnapshotCache, ParquetCache])
@pytest.mark.parametrize("cross_source_caching", [True, False])
def test_complex_cache(cls, cross_source_caching, tmp_path):
    tail = "Tail_652_1"
    flight = "652200101120916"

    con = xo.connect()
    cache_con = xo.connect() if cross_source_caching else con
    cache = cls(source=cache_con, path=tmp_path)

    expr = asof_join_flight_data(con, tail, flight)
    cached = expr.cache(cache=cache)
    assert not cache.exists(expr)
    out = cached.count().execute()
    assert out == 44260
    assert cached.ls.exists()
    assert cache.exists(cached)
    # ParquetCache has an issue with this, regardless of cross_source_caching
    assert cache.exists(expr)



import xorq.vendor.ibis as ibis
from xorq.ibis_yaml.tests.conftest import get_dtype_yaml


def test_add(compiler):
    lit1 = ibis.literal(5)
    lit2 = ibis.literal(3)
    expr = lit1 + lit2

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Add"
    assert expression["left"]["op"] == "Literal"
    assert expression["left"]["value"] == 5
    assert expression["right"]["op"] == "Literal"
    assert expression["right"]["value"] == 3
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    assert dtype_yaml == {
        "op": "DataType",
        "type": "Int8",
        "nullable": {"op": "bool", "value": True},
    }

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_subtract(compiler):
    lit1 = ibis.literal(5)
    lit2 = ibis.literal(3)
    expr = lit1 - lit2

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Subtract"
    assert expression["left"]["op"] == "Literal"
    assert expression["left"]["value"] == 5
    assert expression["right"]["op"] == "Literal"
    assert expression["right"]["value"] == 3
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    assert dtype_yaml == {
        "op": "DataType",
        "type": "Int8",
        "nullable": {"op": "bool", "value": True},
    }

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_multiply(compiler):
    lit1 = ibis.literal(5)
    lit2 = ibis.literal(3)
    expr = lit1 * lit2

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Multiply"
    assert expression["left"]["op"] == "Literal"
    assert expression["left"]["value"] == 5
    assert expression["right"]["op"] == "Literal"
    assert expression["right"]["value"] == 3
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    assert dtype_yaml == {
        "op": "DataType",
        "type": "Int8",
        "nullable": {"op": "bool", "value": True},
    }

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_divide(compiler):
    lit1 = ibis.literal(6.0)
    lit2 = ibis.literal(2.0)
    expr = lit1 / lit2

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Divide"
    assert expression["left"]["op"] == "Literal"
    assert expression["left"]["value"] == 6.0
    assert expression["right"]["op"] == "Literal"
    assert expression["right"]["value"] == 2.0
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    assert dtype_yaml == {
        "op": "DataType",
        "type": "Float64",
        "nullable": {"op": "bool", "value": True},
    }

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_mixed_arithmetic(compiler):
    i = ibis.literal(5)
    f = ibis.literal(2.5)
    expr = i * f

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Multiply"
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    assert dtype_yaml == {
        "op": "DataType",
        "type": "Float64",
        "nullable": {"op": "bool", "value": True},
    }

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_complex_arithmetic(compiler):
    a = ibis.literal(10)
    b = ibis.literal(5)
    c = ibis.literal(2.0)
    expr = (a + b) * c

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]

    assert expression["op"] == "Multiply"
    assert expression["left"]["op"] == "Add"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)



from __future__ import annotations

import decimal
import functools
import math
import operator
from operator import methodcaller

import numpy as np
import pandas as pd
import pytest
from pytest import param

import xorq.api as xo
import xorq.vendor.ibis.expr.datatypes as dt
from xorq.backends.pandas.udf import udf
from xorq.tests.util import assert_series_equal
from xorq.vendor import ibis


@pytest.mark.parametrize(
    "op",
    [
        # comparison
        operator.eq,
        operator.ne,
        operator.lt,
        operator.le,
        operator.gt,
        operator.ge,
    ],
)
def test_binary_operations(t, df, op):
    expr = op(t.plain_float64, t.plain_int64)
    result = expr.execute()
    expected = op(df.plain_float64, df.plain_int64)
    assert_series_equal(result, expected)


@pytest.mark.parametrize("op", [operator.and_, operator.or_, operator.xor])
def test_binary_boolean_operations(t, df, op):
    expr = op(t.plain_int64 == 1, t.plain_int64 == 2)
    result = expr.execute()
    expected = op(df.plain_int64 == 1, df.plain_int64 == 2)
    assert_series_equal(result, expected)


def operate(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except decimal.InvalidOperation:
            return decimal.Decimal("NaN")

    return wrapper


@pytest.mark.parametrize(
    ("ibis_func", "pandas_func"),
    [
        param(methodcaller("round"), round, id="round"),
        param(
            methodcaller("round", 2),
            lambda x: x.quantize(decimal.Decimal(".00")),
            id="round_2",
        ),
        param(
            methodcaller("round", 0),
            lambda x: x.quantize(decimal.Decimal("0.")),
            id="round_0",
        ),
        param(methodcaller("ceil"), lambda x: decimal.Decimal(math.ceil(x)), id="ceil"),
        param(
            methodcaller("floor"), lambda x: decimal.Decimal(math.floor(x)), id="floor"
        ),
        param(
            methodcaller("sign"),
            lambda x: x if not x else decimal.Decimal(1).copy_sign(x),
            id="sign",
        ),
        param(methodcaller("sqrt"), operate(lambda x: x.sqrt()), id="sqrt"),
        param(
            methodcaller("log", 2),
            operate(lambda x: x.ln() / decimal.Decimal(2).ln()),
            id="log_2",
        ),
        param(methodcaller("ln"), operate(lambda x: x.ln()), id="ln"),
        param(
            methodcaller("log2"),
            operate(lambda x: x.ln() / decimal.Decimal(2).ln()),
            id="log2",
        ),
        param(methodcaller("log10"), operate(lambda x: x.log10()), id="log10"),
    ],
)
def test_math_functions_decimal(t, df, ibis_func, pandas_func):
    dtype = dt.Decimal(12, 3)
    context = decimal.Context(prec=dtype.precision)

    def normalize(x):
        x = context.create_decimal(x)
        p = decimal.Decimal(
            f"{'0' * (dtype.precision - dtype.scale)}.{'0' * dtype.scale}"
        )
        return x.quantize(p)

    expr = ibis_func(t.float64_as_strings.cast(dtype))
    result = expr.execute()

    expected = (
        df.float64_as_strings.apply(normalize).apply(pandas_func).apply(normalize)
    )
    assert_series_equal(result, expected.astype(expr.type().to_pandas()))


def test_round_decimal_with_negative_places(t):
    type = dt.Decimal(12, 3)
    expr = t.float64_as_strings.cast(type).round(-1)
    result = expr.execute()
    expected = pd.Series(
        list(map(decimal.Decimal, ["1.0E+2", "2.3E+2", "-1.00E+3"])),
        name="float64_as_strings",
    )
    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("ibis_func", "pandas_func"),
    [
        (lambda x: x.quantile(0), lambda x: x.quantile(0)),
        (lambda x: x.quantile(1), lambda x: x.quantile(1)),
        (lambda x: x.quantile(0.5), lambda x: x.quantile(0.5)),
    ],
)
def test_quantile(t, df, ibis_func, pandas_func):
    result = ibis_func(t.float64_with_zeros).execute()
    expected = pandas_func(df.float64_with_zeros)
    assert result == expected

    assert result == expected

    result = ibis_func(t.int64_with_zeros).execute()
    expected = pandas_func(df.int64_with_zeros)
    assert result == expected


@pytest.mark.parametrize(
    ("ibis_func", "pandas_func"),
    [
        (
            lambda x: x.quantile([0.25, 0.75]),
            lambda x: np.array(x.quantile([0.25, 0.75])),
        )
    ],
)
@pytest.mark.parametrize("column", ["float64_with_zeros", "int64_with_zeros"])
def test_quantile_multi(t, df, ibis_func, pandas_func, column):
    expr = ibis_func(t[column])
    result = expr.execute()
    expected = pandas_func(df[column])
    np.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize(
    ("ibis_func", "exc"),
    [
        # no lower/upper specified
        (lambda x: x.clip(), ValueError),
        # out of range on quantile
        (lambda x: x.quantile(5.0), ValueError),
    ],
)
def test_arraylike_functions_transform_errors(t, ibis_func, exc):
    with pytest.raises(exc):
        ibis_func(t.float64_with_zeros).execute()


def test_quantile_multi_array_access(client, t, df):
    quantile = t.float64_with_zeros.quantile([0.25, 0.5])
    expr = quantile[0], quantile[1]
    result = tuple(map(client.execute, expr))
    expected = tuple(df.float64_with_zeros.quantile([0.25, 0.5]))
    assert result == expected


@pytest.mark.parametrize(
    (
        "left",
        "right",
        "expected_value",
        "expected_type",
        "left_dtype",
        "right_dtype",
    ),
    [
        (True, 1, True, bool, dt.boolean, dt.int64),
        (True, 1.0, True, bool, dt.boolean, dt.float64),
        (True, True, True, bool, dt.boolean, dt.boolean),
        (False, 0, False, bool, dt.boolean, dt.int64),
        (False, 0.0, False, bool, dt.boolean, dt.float64),
        (False, False, False, bool, dt.boolean, dt.boolean),
        (1, True, 1, int, dt.int64, dt.boolean),
        (1, 1.0, 1, int, dt.int64, dt.float64),
        (1, 1, 1, int, dt.int64, dt.int64),
        (0, False, 0, int, dt.int64, dt.boolean),
        (0, 0.0, 0, int, dt.int64, dt.float64),
        (0, 0, 0, int, dt.int64, dt.int64),
        (1.0, True, 1.0, float, dt.float64, dt.boolean),
        (1.0, 1, 1.0, float, dt.float64, dt.int64),
        (1.0, 1.0, 1.0, float, dt.float64, dt.float64),
        (0.0, False, 0.0, float, dt.float64, dt.boolean),
        (0.0, 0, 0.0, float, dt.float64, dt.int64),
        (0.0, 0.0, 0.0, float, dt.float64, dt.float64),
    ],
)
def test_execute_with_same_hash_value_in_scope(
    left, right, expected_value, expected_type, left_dtype, right_dtype
):
    with pytest.warns(FutureWarning, match="v9.0"):

        @udf.elementwise([left_dtype, right_dtype], left_dtype)
        def my_func(x, _):
            return x

    df = pd.DataFrame({"left": [left], "right": [right]})
    con = xo.pandas.connect()
    table = con.from_dataframe(df)

    expr = my_func(table.left, table.right)
    result = con.execute(expr)
    assert isinstance(result, pd.Series)

    result = result.tolist()
    assert result == [expected_value]
    assert type(result[0]) is expected_type


def test_ifelse_returning_bool():
    one = ibis.literal(1)
    two = ibis.literal(2)
    true = ibis.literal(True)
    false = ibis.literal(False)
    expr = ibis.ifelse(one + one == two, true, false)
    result = xo.pandas.connect().execute(expr)
    assert result is True or result is np.True_


@pytest.mark.parametrize(
    ("dtype", "value"),
    [
        pytest.param(dt.float64, 1, id="float_int"),
        pytest.param(dt.float64, True, id="float_bool"),
        pytest.param(dt.int64, 1.0, id="int_float"),
        pytest.param(dt.int64, True, id="int_bool"),
        pytest.param(dt.boolean, 1.0, id="bool_float"),
        pytest.param(dt.boolean, 1, id="bool_int"),
    ],
)
def test_signature_does_not_match_input_type(dtype, value):
    with pytest.warns(FutureWarning, match="v9.0"):

        @udf.elementwise([dtype], dtype)
        def func(x):
            return x

    df = pd.DataFrame({"col": [value]})
    table = xo.pandas.connect().from_dataframe(df)

    result = table.col.execute()
    assert isinstance(result, pd.Series)

    result = result.tolist()
    assert result == [value]
    assert type(result[0]) is type(value)


@pytest.mark.parametrize(
    ("ibis_func", "pandas_func"),
    [
        (
            lambda x: x.approx_median(),
            lambda x: x.median(),
        )
    ],
)
@pytest.mark.parametrize("column", ["float64_with_zeros", "int64_with_zeros"])
def test_approx_median(t, df, ibis_func, pandas_func, column):
    expr = ibis_func(t[column])
    result = expr.execute()
    expected = pandas_func(df[column])
    assert expected == result



import base64
import functools
from functools import wraps
from pathlib import Path
from typing import Any

import cloudpickle
import toolz
from attr import (
    field,
    frozen,
)
from dask.base import tokenize

import xorq.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.caching.strategy import SnapshotStrategy
from xorq.expr.relations import Tag
from xorq.ibis_yaml.config import config
from xorq.ibis_yaml.utils import freeze
from xorq.vendor.ibis.common.collections import FrozenOrderedDict
from xorq.vendor.ibis.expr.schema import Schema


try:
    from enum import StrEnum
except ImportError:
    from strenum import StrEnum


class RefEnum(StrEnum):
    dtype_ref = "dtype_ref"
    node_ref = "node_ref"
    schema_ref = "schema_ref"


class RegistryEnum(StrEnum):
    dtypes = "dtypes"
    nodes = "nodes"
    schemas = "schemas"


FROM_YAML_HANDLERS: dict[str, Any] = {}


def serialize_callable(fn: callable) -> str:
    pickled = cloudpickle.dumps(fn)
    encoded = base64.b64encode(pickled).decode("ascii")
    return encoded


def deserialize_callable(encoded_fn: str) -> callable:
    pickled = base64.b64decode(encoded_fn)
    return cloudpickle.loads(pickled)


class Registry:
    def __init__(self, dtypes=(), nodes=(), schemas=()):
        self.dtypes = dict(dtypes)
        self.nodes = dict(nodes)
        self.schemas = dict(schemas)

    def getstate(self):
        return freeze(
            {
                RegistryEnum.dtypes: self.dtypes,
                RegistryEnum.nodes: self.nodes,
                RegistryEnum.schemas: self.schemas,
            }
        )

    def register_dtype(self, dtype, dtype_dict):
        dtype_ref = f"dtype_{tokenize(dtype_dict)[: config.hash_length]}"
        self.dtypes.setdefault(dtype_ref, dtype_dict)
        frozen = freeze({RefEnum.dtype_ref: dtype_ref})
        return frozen

    def register_node(self, node, node_dict):
        """Register a node and return its name.

        Returns a name like '@read_{hash}', '@filter_{hash}', etc.
        """

        match node:
            case Tag():
                untagged_repr = ("Tag", node.parent.to_expr(), node_dict["metadata"])
                with SnapshotStrategy().normalization_context(node.to_expr()):
                    node_hash = tokenize(untagged_repr)
            case Schema():
                untagged_repr = ("Schema", tuple(node.items()))
                node_hash = tokenize(node)
            case _:
                untagged_repr = node.to_expr().ls.untagged
                with SnapshotStrategy().normalization_context(node.to_expr()):
                    node_hash = tokenize(untagged_repr)
        op_name = node_dict.get("op", "unknown").lower()
        node_ref = f"@{op_name}_{node_hash[: config.hash_length]}"
        node_dict_with_hash = freeze(node_dict | {"snapshot_hash": node_hash})
        if node.__class__.__name__ == "Read" and "memtables" in dict(node.read_kwargs):
            from xorq.common.utils.node_utils import update_read_kwargs

            old_read_kwargs = node_dict_with_hash["read_kwargs"]
            new_path = Path("memtables", dict(old_read_kwargs)["path"].name)
            modified_read_kwargs = update_read_kwargs(
                old_read_kwargs, (("path", new_path),)
            )
            node_dict_with_hash = freeze(
                node_dict_with_hash | {"read_kwargs": modified_read_kwargs}
            )
        self.nodes.setdefault(node_ref, node_dict_with_hash)
        frozen = freeze({RefEnum.node_ref: node_ref})
        return frozen

    def register_schema(self, schema):
        frozen_schema = freeze(
            toolz.valmap(
                functools.partial(translate_to_yaml, context=None),
                schema,
            )
        )
        schema_ref = f"schema_{tokenize(frozen_schema)[: config.hash_length]}"
        self.schemas.setdefault(schema_ref, frozen_schema)
        frozen = freeze({RefEnum.schema_ref: schema_ref})
        return frozen

    def get(self, which, ref):
        match which:
            case RegistryEnum.dtypes:
                dct = self.dtypes
            case RegistryEnum.nodes:
                dct = self.nodes
            case RegistryEnum.schemas:
                dct = self.schemas
            case _:
                raise ValueError(f"don't know how to handle which={which}")
        try:
            return freeze(dct[ref])
        except KeyError as err:
            raise ValueError(
                f"ref {ref} not found in definitions for which={which}"
            ) from err


def _is_absolute_path(instance, attribute, value):
    if value and not Path(value).is_absolute():
        raise ValueError("cache_dir must be absolute")


@frozen
class TranslationContext:
    registry: Registry = field(factory=Registry)
    profiles: FrozenOrderedDict = field(factory=FrozenOrderedDict)
    cache_dir: Path = field(
        default=None,
        converter=toolz.excepts(TypeError, Path),
        validator=_is_absolute_path,
    )

    @property
    def definitions(self):
        return self.registry.getstate()

    def translate_from_yaml(self, yaml_dict: dict) -> Any:
        return translate_from_yaml(yaml_dict, self)

    def translate_to_yaml(self, op: Any) -> dict:
        return safe_translate_to_yaml(op, self)

    def register(self, which, op, frozen=None):
        match which:
            case RegistryEnum.dtypes:
                return self.registry.register_dtype(op, frozen)
            case RegistryEnum.nodes:
                return self.registry.register_node(op, frozen)
            case RegistryEnum.schemas:
                return self.registry.register_schema(op)
            case _:
                raise ValueError(f"don't know how to register {which}")

    def get_definition(self, which, ref):
        return self.registry.get(which, ref)

    def get_dtype(self, dtype_ref):
        dtype_def = self.get_definition(RegistryEnum.dtypes, dtype_ref)
        return self.translate_from_yaml(dtype_def)

    def get_node(self, node_ref):
        node_def = self.get_definition(RegistryEnum.nodes, node_ref)
        return self.translate_from_yaml(node_def)

    def get_schema(self, schema_ref):
        schema_def = self.get_definition(RegistryEnum.schemas, schema_ref)
        schema = Schema(toolz.valmap(self.translate_from_yaml, schema_def))
        return schema


def register_from_yaml_handler(*op_names: str):
    def decorator(func):
        for name in op_names:
            FROM_YAML_HANDLERS[name] = func
        return func

    return decorator


def default_handler(yaml_dict: dict, context: TranslationContext):
    spec = dict(yaml_dict)
    cls = getattr(ops, spec["op"])
    return cls(
        **{
            key: translate_from_yaml(spec[key], context)
            for key in tuple(cls.__signature__.parameters)
        }
    ).to_expr()


@functools.cache
@functools.singledispatch
def translate_from_yaml(yaml_dict: dict, context: TranslationContext) -> Any:
    match yaml_dict:
        case None:
            return None
        case {RefEnum.dtype_ref: dtype_ref, **rest}:
            if rest:
                raise ValueError(
                    f"don't know how to handle additional keys ({tuple(rest)}"
                )
            return context.get_dtype(dtype_ref)
        case {RefEnum.node_ref: node_ref, RefEnum.schema_ref: _schema_ref, **rest}:
            if rest:
                raise ValueError(
                    f"don't know how to handle additional keys ({tuple(rest)}"
                )
            return context.get_node(node_ref)
        case {RefEnum.node_ref: node_ref, **rest}:
            if rest:
                raise ValueError(
                    f"don't know how to handle additional keys ({tuple(rest)}"
                )
            return context.get_node(node_ref)
        case {"op": op_type}:
            return FROM_YAML_HANDLERS.get(op_type, default_handler)(yaml_dict, context)
        case {RefEnum.schema_ref: schema_ref, **rest}:
            if rest:
                raise ValueError(
                    f"don't know how to handle additional keys ({tuple(rest)}"
                )
            return context.get_schema(schema_ref)
        case _:
            raise ValueError(f"don't know how to handle keys ({tuple(yaml_dict)})")


@functools.lru_cache(maxsize=None, typed=True)
@functools.singledispatch
def translate_to_yaml(op: Any, context: TranslationContext) -> dict:
    raise NotImplementedError(f"No translation rule for {type(op)}")


@wraps(translate_to_yaml)
def safe_translate_to_yaml(*args, **kwargs):
    try:
        return translate_to_yaml(*args, **kwargs)
    except Exception:
        return translate_to_yaml.__wrapped__(*args, **kwargs)


@functools.singledispatch
def _translate_type(dtype: dt.DataType) -> dict:
    return freeze({"name": type(dtype).__name__, "nullable": dtype.nullable})



from __future__ import annotations

import sqlglot as sg
import sqlglot.expressions as sge

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis.backends.sql.compilers.base import NULL, STAR, SQLGlotCompiler
from xorq.vendor.ibis.backends.sql.datatypes import ExasolType
from xorq.vendor.ibis.backends.sql.dialects import Exasol
from xorq.vendor.ibis.backends.sql.rewrites import (
    exclude_unsupported_window_frame_from_ops,
    exclude_unsupported_window_frame_from_rank,
    exclude_unsupported_window_frame_from_row_number,
    rewrite_empty_order_by_window,
)


class ExasolCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = Exasol
    type_mapper = ExasolType
    rewrites = (
        exclude_unsupported_window_frame_from_ops,
        exclude_unsupported_window_frame_from_rank,
        exclude_unsupported_window_frame_from_row_number,
        rewrite_empty_order_by_window,
        *SQLGlotCompiler.rewrites,
    )

    UNSUPPORTED_OPS = (
        ops.AnalyticVectorizedUDF,
        ops.ArgMax,
        ops.ArgMin,
        ops.ArrayDistinct,
        ops.ArrayFilter,
        ops.ArrayFlatten,
        ops.ArrayIntersect,
        ops.ArrayMap,
        ops.ArraySort,
        ops.ArrayStringJoin,
        ops.ArrayUnion,
        ops.ArrayZip,
        ops.BitwiseNot,
        ops.CumeDist,
        ops.DateAdd,
        ops.DateSub,
        ops.DateFromYMD,
        ops.DayOfWeekIndex,
        ops.ElementWiseVectorizedUDF,
        ops.IntervalFromInteger,
        ops.IsInf,
        ops.IsNan,
        ops.Levenshtein,
        ops.MultiQuantile,
        ops.RandomUUID,
        ops.ReductionVectorizedUDF,
        ops.RegexExtract,
        ops.RegexReplace,
        ops.RegexSearch,
        ops.RegexSplit,
        ops.RowID,
        ops.Strftime,
        ops.StringJoin,
        ops.StringSplit,
        ops.StringToDate,
        ops.StringToTimestamp,
        ops.TimeDelta,
        ops.TimestampAdd,
        ops.TimestampBucket,
        ops.TimestampDelta,
        ops.TimestampDiff,
        ops.TimestampSub,
        ops.TypeOf,
        ops.Unnest,
    )

    SIMPLE_OPS = {
        ops.Log10: "log10",
        ops.All: "min",
        ops.Any: "max",
    }

    @staticmethod
    def _minimize_spec(start, end, spec):
        if (
            start is None
            and isinstance(getattr(end, "value", None), ops.Literal)
            and end.value.value == 0
            and end.following
        ):
            return None
        return spec

    @staticmethod
    def _gen_valid_name(name: str) -> str:
        """Exasol does not allow dots in quoted column names."""
        return name.replace(".", "_")

    def visit_Modulus(self, op, *, left, right):
        return self.f.anon.mod(left, right)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_date():
            return self.cast(value.isoformat(), dtype)
        elif dtype.is_timestamp():
            val = value.replace(tzinfo=None).isoformat(sep=" ", timespec="milliseconds")
            return self.cast(val, dtype)
        elif dtype.is_array() or dtype.is_struct() or dtype.is_map():
            raise com.UnsupportedBackendType(
                f"{type(dtype).__name__}s are not supported in Exasol"
            )
        elif dtype.is_uuid():
            return sge.convert(str(value))
        return super().visit_NonNullLiteral(op, value=value, dtype=dtype)

    def visit_Date(self, op, *, arg):
        return self.cast(arg, dt.date)

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "sample":
            raise com.UnsupportedOperationError(
                "Exasol only implements `pop` correlation coefficient"
            )

        if (left_type := op.left.dtype).is_boolean():
            left = self.cast(left, dt.Int32(nullable=left_type.nullable))

        if (right_type := op.right.dtype).is_boolean():
            right = self.cast(right, dt.Int32(nullable=right_type.nullable))

        return self.agg.corr(left, right, where=where)

    def visit_GroupConcat(self, op, *, arg, sep, where, order_by):
        if where is not None:
            arg = self.if_(where, arg, NULL)

        if order_by:
            arg = sge.Order(this=arg, expressions=order_by)

        return sge.GroupConcat(this=arg, separator=sep)

    def visit_First(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the exasol backend"
            )
        return self.agg.first_value(arg, where=where, order_by=order_by)

    def visit_Last(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the exasol backend"
            )
        return self.agg.last_value(arg, where=where, order_by=order_by)

    def visit_StartsWith(self, op, *, arg, start):
        return self.f.left(arg, self.f.length(start)).eq(start)

    def visit_EndsWith(self, op, *, arg, end):
        return self.f.right(arg, self.f.length(end)).eq(end)

    def visit_StringFind(self, op, *, arg, substr, start, end):
        return self.f.locate(substr, arg, (start if start is not None else 0) + 1)

    def visit_StringSQLILike(self, op, *, arg, pattern, escape):
        return self.f.upper(arg).like(self.f.upper(pattern))

    def visit_StringContains(self, op, *, haystack, needle):
        return self.f.locate(needle, haystack) > 0

    def visit_ExtractSecond(self, op, *, arg):
        return self.cast(self.f.floor(self.f.extract(self.v.second, arg)), op.dtype)

    def visit_ExtractMillisecond(self, op, *, arg):
        return self.cast(
            (
                self.f.extract(self.v.second, arg)
                - self.f.floor(self.f.extract(self.v.second, arg))
            )
            * 1000,
            op.dtype,
        )

    def visit_ExtractEpochSeconds(self, op, *, arg):
        return self.f.floor(self.f.posix_time(self.cast(arg, dt.timestamp)))

    def visit_StringConcat(self, op, *, arg):
        any_args_null = (a.is_(NULL) for a in arg)
        return self.if_(sg.or_(*any_args_null), NULL, self.f.concat(*arg))

    def visit_CountDistinct(self, op, *, arg, where):
        if where is not None:
            arg = self.if_(where, arg, NULL)
        return self.f.count(sge.Distinct(expressions=[arg]))

    def visit_CountStar(self, op, *, arg, where):
        if where is not None:
            return self.f.sum(self.cast(where, op.dtype))
        return self.f.count(STAR)

    def visit_CountDistinctStar(self, op, *, arg, where):
        cols = [sg.column(k, quoted=self.quoted) for k in op.arg.schema.keys()]
        if where is not None:
            cols = [self.if_(where, c, NULL) for c in cols]
        row = sge.Tuple(expressions=cols)
        return self.f.count(sge.Distinct(expressions=[row]))

    def visit_Median(self, op, *, arg, where):
        return self.visit_Quantile(op, arg=arg, quantile=sge.convert(0.5), where=where)

    visit_ApproxMedian = visit_Median

    def visit_Quantile(self, op, *, arg, quantile, where):
        suffix = "cont" if op.arg.dtype.is_numeric() else "disc"
        funcname = f"percentile_{suffix}"
        if where is not None:
            arg = self.if_(where, arg, NULL)
        return sge.WithinGroup(
            this=self.f[funcname](quantile),
            expression=sge.Order(expressions=[sge.Ordered(this=arg)]),
        )

    visit_ApproxQuantile = visit_Quantile

    def visit_TimestampTruncate(self, op, *, arg, unit):
        short_name = unit.short
        unit_mapping = {"W": "IW"}
        unsupported = {"ms", "us"}

        if short_name in unsupported:
            raise com.UnsupportedOperationError(
                f"Unsupported truncate unit {short_name}"
            )

        if short_name not in unit_mapping:
            return super().visit_TimestampTruncate(op, arg=arg, unit=unit)

        return self.f.date_trunc(unit_mapping[short_name], arg)

    def visit_DateTruncate(self, op, *, arg, unit):
        return self.visit_TimestampTruncate(op, arg=arg, unit=unit)

    def visit_DateDelta(self, op, *, part, left, right):
        # Note: general delta handling could be done based on part (unit),
        #       consider adapting this while implementing time based deltas.
        #       * part = day -> days_between
        #       * part = hour -> hours_between
        #       * ...
        return self.f.days_between(left, right)

    def visit_ExtractDayOfYear(self, op, *, arg):
        return self.cast(self.f.to_char(arg, "DDD"), op.dtype)

    def visit_ExtractWeekOfYear(self, op, *, arg):
        return self.cast(self.f.to_char(arg, "IW"), op.dtype)

    def visit_ExtractIsoYear(self, op, *, arg):
        return self.cast(self.f.to_char(arg, "IYYY"), op.dtype)

    def visit_DayOfWeekName(self, op, *, arg):
        return self.f.concat(
            self.f.substr(self.f.to_char(arg, "DAY"), 0, 1),
            self.f.trim(self.f.lower(self.f.substr(self.f.to_char(arg, "DAY"), 2))),
        )

    def visit_ExtractQuarter(self, op, *, arg):
        return self.cast(self.f.to_char(arg, "Q"), op.dtype)

    def visit_HexDigest(self, op, *, arg, how):
        ibis2exasol = {
            "md5": "hash_md5",
            "sha1": "hash_sha[1]",
            "sha256": "hash_sha256",
            "sha512": "hash_sha512",
        }
        how = how.lower()
        if how not in ibis2exasol:
            raise com.UnsupportedOperationError(
                f"Unsupported hashing algorithm ({how})"
            )
        func = self.f[ibis2exasol[how]]
        return func(arg)

    def visit_BitwiseLeftShift(self, op, *, left, right):
        return self.cast(self.f.bit_lshift(left, right), op.dtype)

    def visit_BitwiseRightShift(self, op, *, left, right):
        return self.cast(self.f.bit_rshift(left, right), op.dtype)

    def visit_BitwiseAnd(self, op, *, left, right):
        return self.cast(self.f.bit_and(left, right), op.dtype)

    def visit_BitwiseOr(self, op, *, left, right):
        return self.cast(self.f.bit_or(left, right), op.dtype)

    def visit_BitwiseXor(self, op, *, left, right):
        return self.cast(self.f.bit_xor(left, right), op.dtype)


compiler = ExasolCompiler()



from __future__ import annotations

import datetime

import pandas as pd
import pytest
from pytest import param

import xorq.api as xo
from xorq.tests.util import assert_frame_equal, assert_series_equal


@pytest.fixture(scope="session")
def con():
    return xo.connect()


@pytest.fixture(scope="session")
def alltypes(con, parquet_dir):
    return con.read_parquet(
        parquet_dir / "functional_alltypes.parquet", table_name="alltypes"
    )


@pytest.mark.parametrize(
    "input_ts",
    [
        param("2023-01-15 14:30:45", id="timestamp"),
        param("2023-12-31 23:59:59", id="year_end"),
    ],
)
def test_date_extraction(compiler, con, input_ts):
    ts = xo.timestamp(input_ts)
    expr = ts.date()

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)
    expected = con.execute(expr)

    assert result == expected
    assert isinstance(result, datetime.date)


def test_date_extraction_column(compiler, con, alltypes):
    expr = alltypes.timestamp_col.date()

    yaml_dict = compiler.to_yaml(expr)
    profiles = {con._profile.hash_name: con}
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    result = roundtrip_expr.execute()
    expected = expr.execute()

    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    "input_ts",
    [
        param("2023-01-15 14:30:45", id="afternoon"),
        param("2023-12-31 00:00:01", id="midnight"),
    ],
)
def test_time_extraction(compiler, con, input_ts):
    ts = xo.timestamp(input_ts)
    expr = ts.time()

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)
    expected = con.execute(expr)

    assert result == expected
    assert isinstance(result, datetime.time)


def test_time_extraction_column(compiler, con, alltypes):
    expr = alltypes.timestamp_col.time()

    yaml_dict = compiler.to_yaml(expr)
    profiles = {con._profile.hash_name: con}
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    result = roundtrip_expr.execute()
    expected = expr.execute()

    assert_series_equal(result, expected)


def test_timestamp_now(compiler, con):
    expr = xo.now()

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)

    assert isinstance(result, datetime.datetime)


def test_date_now(compiler, con):
    expr = xo.today()

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)
    expected = datetime.date.today()

    assert result == expected
    assert isinstance(result, datetime.date)


@pytest.mark.parametrize(
    ("date_str", "interval_days", "expected_str"),
    [
        param("2023-01-15", 10, "2023-01-25", id="add_days"),
        param("2023-12-25", 7, "2024-01-01", id="year_boundary"),
        param("2023-02-28", 1, "2023-03-01", id="month_boundary"),
    ],
)
def test_date_add(compiler, con, date_str, interval_days, expected_str):
    date = xo.literal(date_str).cast("date")
    interval = xo.interval(days=interval_days)
    expr = date + interval

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)
    expected = datetime.date.fromisoformat(expected_str)

    assert result == expected


def test_date_add_column(compiler, con, alltypes):
    date_col = alltypes.timestamp_col.date()
    interval = xo.interval(days=5)
    expr = date_col + interval

    yaml_dict = compiler.to_yaml(expr)
    profiles = {con._profile.hash_name: con}
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    result = roundtrip_expr.execute()
    expected = expr.execute()

    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("date_str", "interval_days", "expected_str"),
    [
        param("2023-01-25", 10, "2023-01-15", id="sub_days"),
        param("2024-01-01", 7, "2023-12-25", id="year_boundary"),
        param("2023-03-01", 1, "2023-02-28", id="month_boundary"),
    ],
)
def test_date_sub(compiler, con, date_str, interval_days, expected_str):
    date = xo.literal(date_str).cast("date")
    interval = xo.interval(days=interval_days)
    expr = date - interval

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)
    expected = datetime.date.fromisoformat(expected_str)

    assert result == expected


def test_date_sub_column(compiler, con, alltypes):
    date_col = alltypes.timestamp_col.date()
    interval = xo.interval(days=3)
    expr = date_col - interval

    yaml_dict = compiler.to_yaml(expr)
    profiles = {con._profile.hash_name: con}
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    result = roundtrip_expr.execute()
    expected = expr.execute()

    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("date1_str", "date2_str"),
    [
        param("2023-01-25", "2023-01-15", id="same_month"),
        param("2024-01-01", "2023-12-25", id="year_boundary"),
        param("2023-03-01", "2023-02-28", id="month_boundary"),
    ],
)
def test_date_diff(compiler, con, date1_str, date2_str):
    date1 = xo.date(date1_str)
    date2 = xo.date(date2_str)
    expr = date1 - date2

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    assert expr.equals(roundtrip_expr)


def test_date_diff_column(compiler, con, alltypes):
    date1 = alltypes.timestamp_col.date()
    date2 = xo.literal("2010-01-01").cast("date")
    expr = date1 - date2

    yaml_dict = compiler.to_yaml(expr)
    profiles = {con._profile.hash_name: con}
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    assert expr.equals(roundtrip_expr)


@pytest.mark.parametrize(
    ("date1_str", "date2_str", "unit", "expected"),
    [
        param("2023-10-01", "2023-09-01", "month", 1, id="month_diff"),
        param("2023-01-10", "2023-01-01", "day", 9, id="day_diff"),
        param("2024-01-01", "2023-01-01", "year", 1, id="year_diff"),
    ],
)
def test_date_delta(compiler, con, date1_str, date2_str, unit, expected):
    date1 = xo.literal(date1_str).cast("date")
    date2 = xo.literal(date2_str).cast("date")
    expr = date1.delta(date2, unit)

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)

    assert result == expected


def test_date_delta_column(compiler, con, alltypes):
    date1 = alltypes.timestamp_col.date()
    date2 = xo.literal("2010-01-01").cast("date")
    expr = date1.delta(date2, "day")

    yaml_dict = compiler.to_yaml(expr)
    profiles = {con._profile.hash_name: con}
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    result = roundtrip_expr.execute()
    expected = expr.execute()

    assert_series_equal(result, expected)


# Test TimestampDelta operation
@pytest.mark.parametrize(
    ("ts1_str", "ts2_str", "unit", "expected"),
    [
        param("2023-01-01 12:00:00", "2023-01-01 10:00:00", "hour", 2, id="hour_diff"),
        param(
            "2023-01-01 10:30:00", "2023-01-01 10:00:00", "minute", 30, id="minute_diff"
        ),
        param("2023-01-02 00:00:00", "2023-01-01 00:00:00", "day", 1, id="day_diff"),
    ],
)
def test_timestamp_delta(compiler, con, ts1_str, ts2_str, unit, expected):
    ts1 = xo.timestamp(ts1_str)
    ts2 = xo.timestamp(ts2_str)
    expr = ts1.delta(ts2, unit)

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)

    assert result == expected


def test_timestamp_delta_column(compiler, con, alltypes):
    ts1 = alltypes.timestamp_col
    ts2 = xo.timestamp("2010-01-01 00:00:00")
    expr = ts1.delta(ts2, "hour")

    yaml_dict = compiler.to_yaml(expr)
    profiles = {con._profile.hash_name: con}
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    result = roundtrip_expr.execute()
    expected = expr.execute()

    assert_series_equal(result, expected)


# Test TimeDelta operation
@pytest.mark.parametrize(
    ("time1_str", "time2_str", "unit", "expected"),
    [
        param("14:30:00", "12:00:00", "hour", 2, id="hour_diff"),
        param("14:30:00", "14:00:00", "minute", 30, id="minute_diff"),
        param("14:00:30", "14:00:00", "second", 30, id="second_diff"),
    ],
)
def test_time_delta(compiler, con, time1_str, time2_str, unit, expected):
    time1 = xo.time(time1_str)
    time2 = xo.time(time2_str)
    expr = time1.delta(time2, unit)

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)

    assert result == expected


def test_date_extraction_null(compiler, con):
    expr = xo.null(xo.dtype("timestamp")).date()

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)

    assert pd.isna(result)


def test_time_extraction_null(compiler, con):
    expr = xo.null(xo.dtype("timestamp")).time()

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)

    assert pd.isna(result)


def test_date_add_null(compiler, con):
    date = xo.null(xo.dtype("date"))
    interval = xo.interval(days=5)
    expr = date + interval

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)

    assert pd.isna(result)


def test_date_diff_null(compiler, con):
    date1 = xo.null(xo.dtype("date"))
    date2 = xo.literal("2023-01-01").cast("date")
    expr = date1 - date2

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    assert expr.equals(roundtrip_expr)


def test_date_delta_null(compiler, con):
    date1 = xo.null(xo.dtype("date"))
    date2 = xo.literal("2023-01-01").cast("date")
    expr = date1.delta(date2, "day")

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    result = con.execute(roundtrip_expr)

    assert pd.isna(result)


# Test with table operations
def test_date_operations_in_projection(compiler, con, alltypes):
    expr = alltypes.select(
        id=alltypes.id,
        date=alltypes.timestamp_col.date(),
        time=alltypes.timestamp_col.time(),
        date_plus_5=alltypes.timestamp_col.date() + xo.interval(days=5),
    ).limit(5)

    yaml_dict = compiler.to_yaml(expr)
    profiles = {con._profile.hash_name: con}
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    result = roundtrip_expr.execute().reset_index(drop=True)
    expected = expr.execute().reset_index(drop=True)

    assert_frame_equal(result, expected)


def test_date_operations_in_filter(compiler, con, alltypes):
    target_date = xo.literal("2010-03-01").cast("date")
    expr = (
        alltypes.filter(alltypes.timestamp_col.date() >= target_date)
        .select("id", "timestamp_col")
        .limit(5)
    )

    yaml_dict = compiler.to_yaml(expr)
    profiles = {con._profile.hash_name: con}
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    result = roundtrip_expr.execute().reset_index(drop=True)
    expected = expr.execute().reset_index(drop=True)

    assert_frame_equal(result, expected)


def test_date_operations_in_aggregation(compiler, con, alltypes):
    expr = (
        alltypes.order_by(alltypes.id, alltypes.int_col)
        .limit(20)
        .group_by(date=alltypes.timestamp_col.date())
        .aggregate(count=alltypes.id.count())
    )

    # Test YAML round-trip
    yaml_dict = compiler.to_yaml(expr)
    profiles = {con._profile.hash_name: con}
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles)

    actual, expected = (
        e.execute().sort_values(e.columns).reset_index(drop=True)
        for e in (roundtrip_expr, expr)
    )

    assert_frame_equal(actual, expected, check_dtype=False)



import datetime
import functools
import operator
import threading
import time

import pandas as pd
import pyarrow as pa
import pytest

import xorq.api as xo
import xorq.expr.datatypes as dt
import xorq.flight.action as A
import xorq.flight.exchanger as E
from xorq.common.utils import classproperty
from xorq.common.utils.rbr_utils import instrument_reader
from xorq.common.utils.tls_utils import TLSKwargs
from xorq.flight import (
    Backend,
    BasicAuth,
    FlightServer,
    FlightUrl,
)
from xorq.flight.action import AddExchangeAction
from xorq.flight.exchanger import EchoExchanger, PandasUDFExchanger
from xorq.flight.tests.conftest import do_agg, field_name, my_udf, return_type
from xorq.loader import load_backend
from xorq.tests.util import assert_frame_equal


def make_flight_url(port, scheme="grpc", auth=None):
    if port is not None:
        assert not FlightUrl.port_in_use(port), f"Port {port} already in use"
    flight_url = (
        FlightUrl(port=port, scheme=scheme)
        if auth is None
        else FlightUrl(
            port=port, scheme=scheme, username=auth.username, password=auth.password
        )
    )
    assert FlightUrl.port_in_use(flight_url.port), (
        f"Port {flight_url.port} should be in use"
    )
    return flight_url


def make_connection_from_name(backend_name):
    fun = load_backend(backend_name).connect if backend_name else xo.connect
    return (
        functools.partial(fun, check_same_thread=False)
        if backend_name == "sqlite"
        else fun
    )


@pytest.mark.parametrize(
    "connection,port",
    [
        pytest.param("duckdb", 5005, id="duckdb"),
        pytest.param("datafusion", 5005, id="datafusion"),
        pytest.param("sqlite", 5005, id="sqlite"),
        pytest.param("", 5005, id="xorq"),
    ],
)
def test_port_in_use(connection, port):
    assert port is not None
    flight_url = make_flight_url(port)
    with pytest.raises(OSError, match="Address already in use"):
        with FlightServer(
            flight_url=flight_url,
            make_connection=make_connection_from_name(connection),
        ) as _:
            # entering the above context releases the port
            # so we won't raise until we enter the second context and try to use it
            flight_url2 = FlightUrl(port=port)  # noqa: F841


class Answer42Action(A.AbstractAction):
    @classproperty
    def name(cls):
        return "answer-42"

    @classmethod
    def description(cls):
        return (
            "the answer to the ultimate question of life, the universe, and everything"
        )

    @classmethod
    def do_action(cls, server, context, action):
        yield A.make_flight_result(42)


def test_list_actions():
    with FlightServer() as flight_server:
        actions = flight_server.client.do_action_one(A.ListActionsAction.name)
        assert actions == tuple(A.actions)


def test_add_action():
    with FlightServer() as flight_server:
        actions = flight_server.client.do_action_one(A.ListActionsAction.name)
        assert Answer42Action.name not in actions
        flight_server.client.do_action(A.AddActionAction.name, Answer42Action)
        actions = flight_server.client.do_action_one(A.ListActionsAction.name)
        assert Answer42Action.name in actions


def test_list_exchanges():
    with FlightServer() as flight_server:
        exchanges = flight_server.client.do_action_one(A.ListExchangesAction.name)
        assert exchanges == tuple(E.exchangers)


@pytest.mark.parametrize(
    "connection,port",
    [
        pytest.param("duckdb", None, id="duckdb"),
        pytest.param("datafusion", None, id="datafusion"),
        pytest.param("sqlite", None, id="sqlite"),
        pytest.param("", None, id="xorq"),
    ],
)
def test_register_and_list_tables(connection, port):
    flight_url = make_flight_url(port)

    with FlightServer(
        flight_url=flight_url,
        verify_client=False,
        make_connection=make_connection_from_name(connection),
    ) as main:
        con = main.con
        assert con.version is not None

        data = pa.table(
            {"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]}
        ).to_pandas()

        con.register(data, table_name="users")
        t = con.table("users")
        actual = xo.execute(t)

        assert t.schema() is not None
        assert "users" in con.list_tables()
        assert isinstance(actual, pd.DataFrame)


@pytest.mark.parametrize("auth", [None, BasicAuth("username", "password")])
@pytest.mark.parametrize("verify_client", [False, True])
def test_tls_encryption(auth, verify_client, tls_kwargs, mtls_kwargs):
    flight_url = make_flight_url(None, scheme="grpc+tls")

    tls_kwargs = TLSKwargs.from_common_name(verify_client=verify_client)

    with FlightServer(
        flight_url=flight_url,
        verify_client=verify_client,
        auth=auth,
        **tls_kwargs.server_kwargs,
    ) as main:
        con = main.con
        assert con.version is not None

        data = pa.table(
            {"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]}
        ).to_pandas()

        con.register(data, table_name="users")
        t = con.table("users")
        actual = xo.execute(t)

        assert t.schema() is not None
        assert "users" in con.list_tables()
        assert isinstance(actual, pd.DataFrame)


def test_failed_auth(tls_kwargs):
    flight_url = make_flight_url(None, scheme="grpc+tls")

    kwargs = TLSKwargs.from_common_name().server_kwargs

    with FlightServer(
        flight_url=flight_url,
        auth=BasicAuth("username", "password"),
        **kwargs,
    ) as server:
        kwargs = {
            "host": server.flight_url.host,
            "port": server.flight_url.port,
            "username": server.auth.username,
            "password": "not_the_password",
            "tls_root_certs": kwargs["root_certificates"],
        }

        from pyarrow._flight import FlightUnauthenticatedError

        with pytest.raises(FlightUnauthenticatedError):
            instance = Backend()
            instance.do_connect(**kwargs)


@pytest.mark.parametrize(
    "connection,port",
    [
        pytest.param("duckdb", None, id="duckdb"),
        pytest.param("datafusion", None, id="datafusion"),
        pytest.param("sqlite", None, id="sqlite"),
        pytest.param("", None, id="xorq"),
    ],
)
def test_into_backend_flight_server(connection, port, parquet_dir):
    batting = xo.read_parquet(parquet_dir / "batting.parquet")
    flight_url = make_flight_url(port)

    with FlightServer(
        flight_url=flight_url,
        verify_client=False,
        make_connection=make_connection_from_name(connection),
    ) as main:
        con = main.con
        t = batting.filter(batting.yearID == 2015).into_backend(con, "xo_batting")
        expr = (
            t.join(t, "playerID")
            .limit(15)
            .select(player_id="playerID", year_id="yearID_right")
        )

        assert not expr.execute().empty


@pytest.mark.parametrize(
    "connection,port",
    [
        pytest.param("duckdb", None, id="duckdb"),
        pytest.param("datafusion", None, id="datafusion"),
        pytest.param("sqlite", None, id="sqlite"),
        pytest.param("", None, id="xorq"),
    ],
)
def test_read_parquet(connection, port, parquet_dir):
    flight_url = make_flight_url(port)
    with FlightServer(
        flight_url=flight_url,
        verify_client=False,
        make_connection=make_connection_from_name(connection),
    ) as main:
        con = main.con
        batting = con.read_parquet(parquet_dir / "batting.parquet")
        assert xo.execute(batting) is not None


@pytest.mark.parametrize(
    "connection,port",
    [
        pytest.param("duckdb", None, id="duckdb"),
        pytest.param("datafusion", None, id="datafusion"),
        pytest.param("sqlite", None, id="sqlite"),
        pytest.param("", None, id="xorq"),
    ],
)
def test_exchange(connection, port):
    flight_url = make_flight_url(port)

    def my_f(df):
        return df[["a", "b"]].sum(axis=1)

    with FlightServer(
        flight_url=flight_url,
        verify_client=False,
        make_connection=make_connection_from_name(connection),
    ) as main:
        client = main.client
        udf_exchanger = PandasUDFExchanger(
            my_f,
            schema_in=xo.schema({"a": int, "b": int}),
            name="x",
            typ=dt.int64,
            append=True,
        )
        client.do_action(AddExchangeAction.name, udf_exchanger, options=client._options)

        # a small example
        df_in = pd.DataFrame({"a": [1], "b": [2], "c": [100]})
        fut, rbr = client.do_exchange(
            udf_exchanger.command,
            xo.memtable(df_in),
        )
        df_out = rbr.read_pandas()
        writes_reads = fut.result()

        assert writes_reads["n_writes"] == writes_reads["n_reads"]
        assert df_out is not None

        # demonstrate streaming
        df_in = pd.DataFrame(
            {
                "a": range(100_000),
                "b": range(100_000, 200_000),
                "c": range(200_000, 300_000),
            }
        )
        fut, rbr = client.do_exchange_batches(
            udf_exchanger.command,
            instrument_reader(pa.Table.from_pandas(df_in).to_reader(max_chunksize=100)),
        )
        first_batch = next(rbr)
        first_batch_time = datetime.datetime.now()
        assert first_batch is not None, f"must get first batch by {first_batch_time}"

        rest = rbr.read_pandas()
        rest_time = datetime.datetime.now()
        assert rest is not None, f"must get first batch by {rest_time}"
        assert first_batch_time < rest_time

        writes_reads = fut.result()
        assert writes_reads["n_writes"] == 1000  # because
        assert writes_reads["n_reads"] == 1000


@pytest.mark.parametrize(
    "connection",
    (
        "duckdb",
        "datafusion",
        "sqlite",
    ),
)
def test_reentry(connection):
    df_in = pd.DataFrame({"a": [1], "b": [2], "c": [100]})
    with FlightServer(
        verify_client=False,
        make_connection=make_connection_from_name(connection),
    ) as server:
        fut, rbr = server.client.do_exchange_batches(
            EchoExchanger.command,
            pa.RecordBatchReader.from_stream(df_in),
        )
        df_out = rbr.read_pandas()
        assert df_in.equals(df_out)
    with server:
        fut, rbr = server.client.do_exchange_batches(
            EchoExchanger.command,
            pa.RecordBatchReader.from_stream(df_in),
        )
        df_out = rbr.read_pandas()
        assert df_in.equals(df_out)


@pytest.mark.parametrize(
    "connection",
    (
        "duckdb",
        "datafusion",
        "sqlite",
    ),
)
def test_serve_close(connection):
    df_in = pd.DataFrame({"a": [1], "b": [2], "c": [100]})
    server = FlightServer(
        verify_client=False,
        make_connection=make_connection_from_name(connection),
    )

    server.serve()
    fut, rbr = server.client.do_exchange_batches(
        EchoExchanger.command,
        pa.RecordBatchReader.from_stream(df_in),
    )
    df_out = rbr.read_pandas()
    assert df_in.equals(df_out)
    server.close()

    server.serve()
    fut, rbr = server.client.do_exchange_batches(
        EchoExchanger.command,
        pa.RecordBatchReader.from_stream(df_in),
    )
    df_out = rbr.read_pandas()
    assert df_in.equals(df_out)
    server.close()


def test_ctor_exchanger_registration():
    def dummy(df: pd.DataFrame):
        return pd.DataFrame({"row_count": [42]})

    schema_in = xo.schema({"dummy": "int64"})
    dummy_udxf = E.make_udxf(
        dummy,
        schema_in,
        xo.schema({"row_count": "int64"}),
    )
    flight_server = FlightServer(exchangers=[dummy_udxf])
    with flight_server:
        client = flight_server.client
        available = client.do_action_one(A.ListExchangesAction.name)
        assert dummy_udxf.command in available
        client.do_exchange(
            dummy_udxf.command,
            xo.memtable({"dummy": [0]}, schema=schema_in),
        )


@pytest.mark.parametrize(
    "expr",
    [
        pytest.param(xo.table({"id": int}, name="users").count(), id="scalar"),
        pytest.param(xo.literal(1), id="literal"),
        pytest.param(xo.table({"id": int}, name="users").id, id="column"),
    ],
)
def test_execute_query_non_relation_expr(expr):
    flight_url = make_flight_url(None)
    data = pa.table({"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]}).to_pandas()

    with FlightServer(
        flight_url=flight_url,
        verify_client=False,
        make_connection=xo.duckdb.connect,
    ) as main:
        main.con.register(data, table_name="users")
        actual = main.client.execute(expr)
        assert isinstance(actual, pa.Table)
        assert len(actual) > 0


@pytest.mark.parametrize("block", [True, False])
def test_server_blocks(block):
    flight_url = make_flight_url(None)
    server = FlightServer(
        flight_url=flight_url,
        verify_client=False,
        make_connection=xo.duckdb.connect,
    )

    server_thread = threading.Thread(
        target=functools.partial(server.serve, block=block), daemon=True
    )

    is_blocking = True

    def check_if_still_running():
        nonlocal is_blocking
        time.sleep(1)
        is_blocking = server_thread.is_alive()

    server_thread.start()

    checker_thread = threading.Thread(target=check_if_still_running)
    checker_thread.start()
    checker_thread.join()

    # Try to stop the inner server
    server.server.shutdown()
    server_thread.join(timeout=2.0)

    assert is_blocking == block
    assert not server_thread.is_alive()


def test_exchange_server_from_udxf(con, diamonds, baseline):
    input_expr = diamonds.pipe(do_agg)
    process_df = operator.methodcaller("assign", **{field_name: my_udf.fn})
    maybe_schema_in = input_expr.schema()
    maybe_schema_out = xo.schema(input_expr.schema() | {field_name: return_type})
    command = "diamonds_exchange_command"
    expr = xo.expr.relations.flight_udxf(
        input_expr,
        process_df=process_df,
        maybe_schema_in=maybe_schema_in,
        maybe_schema_out=maybe_schema_out,
        con=con,
        make_udxf_kwargs={"name": my_udf.__name__, "command": command},
    ).order_by("cut")

    with FlightServer.from_udxf(expr) as server:
        client = server.client
        assert client is not None

        _, rbr = client.do_exchange(
            command,
            input_expr,
        )

        actual = rbr.read_pandas().sort_values("cut", ignore_index=True)
        expected = baseline.sort_values("cut", ignore_index=True)
        assert_frame_equal(
            actual,
            expected,
            check_exact=False,
        )



import hashlib
import operator
import pathlib
import re
from unittest.mock import (
    Mock,
    patch,
)

import dask
import pytest
import toolz

import xorq.api as xo
import xorq.common.utils.dask_normalize  # noqa: F401
from xorq.caching import (
    ParquetCache,
    SourceSnapshotCache,
)
from xorq.common.utils.dask_normalize import (
    get_normalize_token_subset,
)
from xorq.common.utils.dask_normalize.dask_normalize_utils import (
    file_digest,
    gen_batches,
    manual_file_digest,
    patch_normalize_token,
    walk_normalized,
)
from xorq.ibis_yaml.compiler import build_expr


def test_ensure_deterministic():
    assert dask.config.get("tokenize.ensure-deterministic")


def test_unregistered_raises():
    class Unregistered:
        pass

    with pytest.raises(ValueError, match="cannot be deterministically hashed"):
        dask.base.tokenize(Unregistered())


@pytest.fixture(scope="function")
def alltypes_df(pg):
    return pg.table("functional_alltypes").execute()


@pytest.fixture(scope="function")
def batting(pg):
    return pg.table("batting")


@pytest.mark.snapshot_check
def test_tokenize_datafusion_memory_expr(alltypes_df, snapshot):
    con = xo.datafusion.connect()
    t = con.register(alltypes_df, "t")
    f = Mock(side_effect=toolz.functoolz.return_none)
    with patch_normalize_token(type(con), f=f):
        actual = dask.base.tokenize(t)
    f.assert_not_called()
    snapshot.assert_match(actual, "datafusion_memory_key.txt")


@pytest.mark.snapshot_check
def test_tokenize_datafusion_parquet_expr(alltypes_df, tmp_path, snapshot):
    path = pathlib.Path(tmp_path).joinpath("data.parquet")
    alltypes_df.to_parquet(path)
    con = xo.datafusion.connect()
    t = con.register(path, "t")
    # work around tmp_path variation
    (prefix, suffix) = (
        re.escape(part)
        for part in (
            r"file_groups={1 group: [[",
            r"]]",
        )
    )
    to_hash = re.sub(
        prefix + f".*?/{path.name}" + suffix,
        prefix + f"/{path.name}" + suffix,
        str(tuple(dask.base.normalize_token(t))),
    )
    actual = hashlib.md5(to_hash.encode(), usedforsecurity=False).hexdigest()
    snapshot.assert_match(actual, "datafusion_key.txt")


@pytest.mark.snapshot_check
def test_tokenize_pandas_expr(alltypes_df, snapshot):
    con = xo.pandas.connect()
    t = con.create_table("t", alltypes_df)
    f = Mock(side_effect=toolz.functoolz.return_none)
    with patch_normalize_token(type(t.op().source), f=f):
        actual = dask.base.tokenize(t)
    f.assert_not_called()
    snapshot.assert_match(actual, "pandas_key.txt")


@pytest.mark.snapshot_check
def test_tokenize_duckdb_expr(batting, snapshot):
    con = xo.duckdb.connect()
    t = con.create_table("dashed-name", batting.to_pyarrow())
    f = Mock(side_effect=toolz.functoolz.return_none)
    with patch_normalize_token(type(con), f=f):
        actual = dask.base.tokenize(t)
    f.assert_not_called()

    snapshot.assert_match(actual, "duckdb_key.txt")


@pytest.mark.snapshot_check
def test_pandas_snapshot_key(alltypes_df, snapshot):
    con = xo.pandas.connect()
    t = con.create_table("t", alltypes_df)
    cache = SourceSnapshotCache.from_kwargs(source=con)
    actual = cache.strategy.calc_key(t)
    snapshot.assert_match(actual, "pandas_snapshot_key.txt")


@pytest.mark.snapshot_check
def test_duckdb_snapshot_key(batting, snapshot):
    con = xo.duckdb.connect()
    t = con.create_table("dashed-name", batting.to_pyarrow())
    cache = SourceSnapshotCache.from_kwargs(source=con)
    actual = cache.strategy.calc_key(t)
    snapshot.assert_match(actual, "duckdb_snapshot_key.txt")


@pytest.mark.parametrize(
    "target, expected",
    (
        (1, 1),
        (1.0, 1),
        ("two", 2),
        # bytes are converted to text
        (b"three", 0),
        ("three", 3),
    ),
)
def test_walk_normalized(target, expected):
    normalized = (
        [
            (
                b"three",
                b"three",
                1,
                [
                    "two",
                ],
            ),
        ],
        (
            2,
            (
                3,
                b"three",
            ),
            "two",
        ),
    )
    f = toolz.curry(operator.eq, target)
    actual = sum(walk_normalized(f, normalized))
    assert actual == expected


def test_normalize_token_lookup():
    assert not get_normalize_token_subset()


def test_dask_tokenize_object():
    class MissingDaskTokenize:
        pass

    tokenize_value = 1

    class HasDaskTokenize:
        def __dask_tokenize__(self):
            return tokenize_value

    assert dask.base.normalize_token(HasDaskTokenize()) == tokenize_value
    with pytest.raises(ValueError):
        dask.base.tokenize(MissingDaskTokenize())


def test_partitioning():
    path = pathlib.Path(xo.config.options.pins.get_path("batting"))
    assert len(tuple(gen_batches(path))) > 1
    content = b"".join(gen_batches(path))
    assert content == path.read_bytes()


def test_file_digest():
    path = pathlib.Path(xo.config.options.pins.get_path("batting"))
    actual = manual_file_digest(path)
    expected = file_digest(path)
    assert actual == expected


def test_patch_normalize_token():
    def make_type(name):
        cls = type(name, (), {})
        cls.__module__ = name
        return cls

    names = ("to_retain", "to_drop")
    name_to_cls = {name: make_type(name) for name in names}
    name_to_mock = {
        cls.__module__: Mock(
            side_effect=toolz.partial(
                dask.base.normalize_token.register,
                cls,
                toolz.functoolz.return_none,
            ),
        )
        for cls in name_to_cls.values()
    }

    assert not any(
        cls.__module__ in dask.base.normalize_token._lazy
        for cls in name_to_cls.values()
    )
    assert not any(
        cls in dask.base.normalize_token._lookup for cls in name_to_cls.values()
    )
    values = {name: name_to_mock[name] for name in ("to_retain",)}
    with patch.dict(
        dask.base.normalize_token._lazy,
        values=values,
    ):
        with patch_normalize_token(
            name_to_cls["to_drop"],
            f=toolz.functoolz.return_none,
        ):
            assert name_to_cls["to_retain"] not in dask.base.normalize_token._lookup
            for name in names:
                dask.base.tokenize(name_to_cls[name]())
            for cls in name_to_cls.values():
                assert cls in dask.base.normalize_token._lookup
        assert name_to_cls["to_drop"] not in dask.base.normalize_token._lookup
        assert name_to_cls["to_retain"] in dask.base.normalize_token._lookup

    assert "to_retain" not in dask.base.normalize_token._lookup
    assert name_to_cls["to_retain"] in dask.base.normalize_token._lookup


def test_parquet_cache_tokenize_stable_across_cloudpickle():
    import cloudpickle

    con = xo.connect()
    cache = ParquetCache.from_kwargs(source=con)
    token_before = dask.base.tokenize(cache)
    cache2 = cloudpickle.loads(cloudpickle.dumps(cache))
    token_after = dask.base.tokenize(cache2)
    assert token_before == token_after


def test_different_cache_types_produce_different_hashes():
    t = xo.memtable({"a": [1, 2, 3], "b": ["x", "y", "z"]})
    c0 = t.cache()
    c1 = t.cache(ParquetCache.from_kwargs())
    b0 = build_expr(c0)
    b1 = build_expr(c1)
    assert b0 != b1



import functools
import itertools

import toolz


count = itertools.count()


def return_constant(value):
    def wrapped(*args, **kwargs):
        return value

    return wrapped


@toolz.curry
def log_excepts(f, exception=Exception):
    # file logger
    # from xorq.common.utils.logging_utils import get_logger

    # print logger
    from structlog import get_logger

    logger = get_logger(__name__)

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        i = next(count)
        try:
            logger.info(f"{f.__name__} :: entering :: {i}")
            value = f(*args, **kwargs)
            logger.info(f"{f.__name__} :: exiting  :: {i}")
            return value
        except exception:
            logger.exception("exception!")

    return wrapper


@toolz.curry
def maybe_log_excepts(f, exception=Exception, debug=None):
    from xorq.config import options

    if options.debug or debug:
        return log_excepts(f, exception=exception)
    else:
        return f


@toolz.curry
def with_lock(lock, f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        with lock:
            return f(*args, **kwargs)

    return wrapper


@toolz.curry
def if_not_none(f, value):
    return value if value is None else f(value)



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
from pytest import param

import xorq.api as xo
import xorq.common.exceptions as com
import xorq.expr.datatypes as dt
from xorq.tests.util import assert_frame_equal, reduction_tolerance
from xorq.vendor.ibis import _
from xorq.vendor.ibis import literal as L


@xo.udf.agg.builtin
def mean_udf(s: dt.double) -> dt.double:
    return s.mean()


aggregate_test_params = [
    param(lambda t: t.double_col.mean(), lambda t: t.double_col.mean(), id="mean"),
    param(lambda t: t.double_col.min(), lambda t: t.double_col.min(), id="min"),
    param(lambda t: t.double_col.max(), lambda t: t.double_col.max(), id="max"),
    param(
        lambda t: (t.double_col + 5).sum(),
        lambda t: (t.double_col + 5).sum(),
        id="complex_sum",
    ),
    param(
        lambda t: t.timestamp_col.max(),
        lambda t: t.timestamp_col.max(),
        id="timestamp_max",
    ),
]


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    aggregate_test_params,
)
def test_aggregate(alltypes, alltypes_df, result_fn, expected_fn):
    expr = alltypes.aggregate(tmp=result_fn)
    result = expr.execute()

    # Create a single-row single-column dataframe with the Pandas `agg` result
    # (to match the output format of Ibis `aggregate`)
    expected = pd.DataFrame({"tmp": [expected_fn(alltypes_df)]})

    assert_frame_equal(result, expected, check_dtype=False)


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    aggregate_test_params,
)
def test_aggregate_grouped(alltypes, alltypes_df, result_fn, expected_fn):
    grouping_key_col = "bigint_col"

    # Two (equivalent) variations:
    #  1) `group_by` then `aggregate`
    #  2) `aggregate` with `by`
    expr1 = alltypes.group_by(grouping_key_col).aggregate(tmp=result_fn)
    expr2 = alltypes.aggregate(tmp=result_fn, by=grouping_key_col)
    result1 = expr1.execute()
    result2 = expr2.execute()

    # Note: Using `reset_index` to get the grouping key as a column
    expected = (
        alltypes_df.groupby(grouping_key_col)
        .apply(expected_fn)
        .rename("tmp")
        .reset_index()
    )

    # Row ordering may differ depending on backend, so sort on the
    # grouping key
    result1 = result1.sort_values(by=grouping_key_col).reset_index(drop=True)
    result2 = result2.sort_values(by=grouping_key_col).reset_index(drop=True)
    expected = expected.sort_values(by=grouping_key_col).reset_index(drop=True)

    assert_frame_equal(result1, expected, check_dtype=False)
    assert_frame_equal(result2, expected, check_dtype=False)


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    [
        param(
            lambda t, where: t.bool_col.count(where=where),
            lambda t, where: len(t.bool_col[where].dropna()),
            id="count",
        ),
        param(
            lambda t, where: t.bool_col.nunique(where=where),
            lambda t, where: t.bool_col[where].dropna().nunique(),
            id="nunique",
        ),
        param(
            lambda t, where: t.bool_col.any(where=where),
            lambda t, where: t.bool_col[where].any(),
            id="any",
        ),
        param(
            lambda t, where: t.bool_col.notany(where=where),
            lambda t, where: ~t.bool_col[where].any(),
            id="notany",
        ),
        param(
            lambda t, where: -t.bool_col.any(where=where),
            lambda t, where: ~t.bool_col[where].any(),
            id="any_negate",
        ),
        param(
            lambda t, where: t.bool_col.all(where=where),
            lambda t, where: t.bool_col[where].all(),
            id="all",
        ),
        param(
            lambda t, where: t.bool_col.notall(where=where),
            lambda t, where: ~t.bool_col[where].all(),
            id="notall",
        ),
        param(
            lambda t, where: -t.bool_col.all(where=where),
            lambda t, where: ~t.bool_col[where].all(),
            id="all_negate",
        ),
        param(
            lambda t, where: t.double_col.sum(where=where),
            lambda t, where: t.double_col[where].sum(),
            id="sum",
        ),
        param(
            lambda t, where: (t.int_col > 0).sum(where=where),
            lambda t, where: (t.int_col > 0)[where].sum(),
            id="bool_sum",
        ),
        param(
            lambda t, where: t.double_col.mean(where=where),
            lambda t, where: t.double_col[where].mean(),
            id="mean",
        ),
        param(
            lambda t, where: t.double_col.min(where=where),
            lambda t, where: t.double_col[where].min(),
            id="min",
        ),
        param(
            lambda t, where: t.double_col.max(where=where),
            lambda t, where: t.double_col[where].max(),
            id="max",
        ),
        param(
            lambda t, where: t.double_col.std(how="sample", where=where),
            lambda t, where: t.double_col[where].std(ddof=1),
            id="std",
        ),
        param(
            lambda t, where: t.double_col.var(how="sample", where=where),
            lambda t, where: t.double_col[where].var(ddof=1),
            id="var",
        ),
        param(
            lambda t, where: t.double_col.std(how="pop", where=where),
            lambda t, where: t.double_col[where].std(ddof=0),
            id="std_pop",
        ),
        param(
            lambda t, where: t.double_col.var(how="pop", where=where),
            lambda t, where: t.double_col[where].var(ddof=0),
            id="var_pop",
        ),
        param(
            lambda t, where: t.string_col.nunique(where=where),
            lambda t, where: t.string_col[where].nunique(),
            id="string_nunique",
        ),
        param(
            lambda t, where: t.double_col.first(where=where),
            lambda t, where: t.double_col[where].iloc[0],
            id="first",
        ),
        param(
            lambda t, where: t.double_col.last(where=where),
            lambda t, where: t.double_col[where].iloc[-1],
            id="last",
        ),
        param(
            lambda t, where: t.bigint_col.bit_and(where=where),
            lambda t, where: np.bitwise_and.reduce(t.bigint_col[where].values),
            id="bit_and",
        ),
        param(
            lambda t, where: t.bigint_col.bit_or(where=where),
            lambda t, where: np.bitwise_or.reduce(t.bigint_col[where].values),
            id="bit_or",
        ),
        param(
            lambda t, where: t.bigint_col.bit_xor(where=where),
            lambda t, where: np.bitwise_xor.reduce(t.bigint_col[where].values),
            id="bit_xor",
        ),
        param(
            lambda t, where: t.count(where=where),
            lambda t, where: len(t[where]),
            id="count_star",
        ),
    ],
)
@pytest.mark.parametrize(
    ("ibis_cond", "pandas_cond"),
    [
        param(lambda _: None, lambda _: slice(None), id="no_cond"),
        param(
            lambda t: t.string_col.isin(["1", "7"]),
            lambda t: t.string_col.isin(["1", "7"]),
            id="is_in",
        ),
        param(
            lambda _: xo._.string_col.isin(["1", "7"]),
            lambda t: t.string_col.isin(["1", "7"]),
            id="is_in_deferred",
        ),
    ],
)
def test_reduction_ops(
    alltypes,
    alltypes_df,
    result_fn,
    expected_fn,
    ibis_cond,
    pandas_cond,
):
    expr = alltypes.agg(tmp=result_fn(alltypes, ibis_cond(alltypes))).tmp
    result = expr.execute().squeeze()
    expected = expected_fn(alltypes_df, pandas_cond(alltypes_df))

    try:
        np.testing.assert_allclose(result, expected, rtol=reduction_tolerance)
    except TypeError:  # assert_allclose only handles numerics
        # if we're not testing numerics, then the arrays should be exactly equal
        np.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    [
        param(
            lambda t, where: t.G.cov(t.RBI, where=where, how="pop"),
            lambda t, where: t.G[where].cov(t.RBI[where], ddof=0),
            id="covar_pop",
        ),
        param(
            lambda t, where: t.G.cov(t.RBI, where=where, how="sample"),
            lambda t, where: t.G[where].cov(t.RBI[where], ddof=1),
            id="covar_samp",
        ),
        param(
            lambda t, where: t.G.corr(t.RBI, where=where, how="pop"),
            lambda t, where: t.G[where].corr(t.RBI[where]),
            id="corr_pop",
        ),
        param(
            lambda t, where: t.G.corr(t.RBI, where=where, how="sample"),
            lambda t, where: t.G[where].corr(t.RBI[where]),
            id="corr_samp",
        ),
        param(
            lambda t, where: (t.G > 34.0).cov(
                t.G <= 34.0,
                where=where,
                how="pop",
            ),
            lambda t, where: (t.G[where] > 34.0).cov(t.G[where] <= 34.0, ddof=0),
            id="covar_pop_bool",
        ),
        param(
            lambda t, where: (t.G > 34.0).corr(
                t.G <= 34.0,
                where=where,
                how="pop",
            ),
            lambda t, where: (t.G[where] > 34.0).corr(t.G[where] <= 34.0),
            id="corr_pop_bool",
        ),
    ],
)
@pytest.mark.parametrize(
    ("ibis_cond", "pandas_cond"),
    [
        param(lambda _: None, lambda _: slice(None), id="no_cond"),
        param(
            lambda t: t.yearID.isin([2009, 2015]),
            lambda t: t.yearID.isin([2009, 2015]),
            id="cond",
        ),
    ],
)
def test_corr_cov(
    batting,
    batting_df,
    result_fn,
    expected_fn,
    ibis_cond,
    pandas_cond,
):
    expr = result_fn(batting, ibis_cond(batting)).name("tmp")
    result = expr.execute()

    expected = expected_fn(batting_df, pandas_cond(batting_df))

    assert pytest.approx(result) == expected


def test_approx_median(alltypes):
    expr = alltypes.double_col.approx_median()
    result = expr.execute()
    assert isinstance(result, float)


def test_median(alltypes, alltypes_df):
    expr = alltypes.double_col.median()
    result = expr.execute()
    expected = alltypes_df.double_col.median()
    assert result == expected


def test_topk_op(alltypes, alltypes_df):
    # TopK expression will order rows by "count" but each backend
    # can have different result for that.
    # Note: Maybe would be good if TopK could order by "count"
    # and the field used by TopK
    t = alltypes.order_by(alltypes.string_col)
    df = alltypes_df.sort_values("string_col")
    expr = t.string_col.topk(3)
    result = expr.execute()
    expected = df.groupby("string_col")["string_col"].count().head(3)
    assert all(result.iloc[:, 1].values == expected.values)


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    [
        param(
            lambda t: t.semi_join(t.string_col.topk(3), "string_col"),
            lambda t: t[
                t.string_col.isin(
                    t.groupby("string_col")["string_col"].count().head(3).index
                )
            ],
            id="string_col_filter_top3",
        )
    ],
)
def test_topk_filter_op(alltypes, alltypes_df, result_fn, expected_fn):
    # TopK expression will order rows by "count" but each backend
    # can have different result for that.
    # Note: Maybe would be good if TopK could order by "count"
    # and the field used by TopK
    t = alltypes.order_by(alltypes.string_col)
    df = alltypes_df.sort_values("string_col")
    expr = result_fn(t)
    result = expr.execute()
    expected = expected_fn(df)
    assert result.shape[0] == expected.shape[0]


def test_binds_are_cast(alltypes):
    expr = alltypes.aggregate(
        high_line_count=(
            alltypes.string_col.case().when("1-URGENT", 1).else_(0).end().sum()
        )
    )

    expr.execute()


def test_agg_sort(alltypes):
    query = alltypes.aggregate(count=alltypes.count())
    with pytest.raises(com.IntegrityError):
        query.order_by(alltypes.year)


def test_filter(alltypes, alltypes_df):
    expr = (
        alltypes[_.string_col == "1"]
        .mutate(x=L(1, "int64"))
        .group_by(_.x)
        .aggregate(sum=_.double_col.sum())
    )

    result = expr.execute().astype({"x": "int64"})
    expected = (
        alltypes_df.loc[alltypes_df.string_col == "1", :]
        .assign(x=1)
        .groupby("x")
        .double_col.sum()
        .rename("sum")
        .reset_index()
    )
    assert_frame_equal(result, expected, check_like=True)


def test_agg_name_in_output_column(alltypes):
    query = alltypes.aggregate([alltypes.int_col.min(), alltypes.int_col.max()])
    df = query.execute()
    assert "min" in df.columns[0].lower()
    assert "max" in df.columns[1].lower()


def test_grouped_case(con):
    table = xo.memtable({"key": [1, 1, 2, 2], "value": [10, 30, 20, 40]})

    case_expr = xo.cases((table.value < 25, table.value), else_=xo.null())

    expr = table.group_by("key").aggregate(mx=case_expr.max()).order_by("key")
    result = con.execute(expr)
    expected = pd.DataFrame({"key": [1, 2], "mx": [10, 20]})
    assert_frame_equal(result, expected)


def test_value_counts_on_expr(alltypes, alltypes_df):
    expr = alltypes.bigint_col.add(1).value_counts()
    columns = expr.columns
    expr = expr.order_by(columns)
    result = expr.execute().sort_values(columns).reset_index(drop=True)
    expected = alltypes_df.bigint_col.add(1).value_counts().reset_index()
    expected.columns = columns
    expected = expected.sort_values(by=columns).reset_index(drop=True)
    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    [
        param(
            lambda t, where: t.double_col.quantile(0.5, where=where),
            lambda t, where: t.double_col[where].quantile(0.5),
            id="quantile",
        ),
    ],
)
@pytest.mark.parametrize(
    ("ibis_cond", "pandas_cond"),
    [
        param(lambda _: None, lambda _: slice(None), id="no_cond"),
        param(
            lambda t: t.string_col.isin(["1", "7"]),
            lambda t: t.string_col.isin(["1", "7"]),
            id="is_in",
        ),
    ],
)
def test_quantile(
    alltypes,
    alltypes_df,
    result_fn,
    expected_fn,
    ibis_cond,
    pandas_cond,
):
    expr = alltypes.agg(tmp=result_fn(alltypes, ibis_cond(alltypes))).tmp
    result = expr.execute().squeeze()
    expected = expected_fn(alltypes_df, pandas_cond(alltypes_df))
    assert pytest.approx(result) == expected


@pytest.mark.parametrize(
    "method,expected",
    [
        pytest.param(lambda col: col.first(order_by="ob"), 4, id="first_asc"),
        pytest.param(lambda col: col.last(order_by="ob"), 5, id="last_asc"),
        pytest.param(
            lambda col: col.first(order_by=xo._.ob.desc()), 5, id="first_desc"
        ),
        pytest.param(lambda col: col.last(order_by=xo._.ob.desc()), 4, id="last_desc"),
    ],
)
def test_first_last_ordered_in_mutate(alltypes, con, method, expected):
    t = alltypes.select(
        a=xo._.tinyint_col, val=xo._.int_col, ob=xo._.bigint_col
    ).filter(((xo._.val == 4) & (xo._.ob == 40)) | ((xo._.val == 5) & (xo._.ob == 50)))
    expr = t.mutate(new=method(t.val)).limit(10)
    actual = con.to_pyarrow(expr.new).to_pylist()
    assert actual == [expected] * 10



import os
import random
import string
import warnings
from pathlib import Path
from typing import Optional

from attr import (
    evolve,
    field,
    frozen,
)
from attr.validators import (
    instance_of,
)
from cryptography.hazmat.primitives.asymmetric import (
    rsa,
)
from cryptography.hazmat.primitives.serialization import (
    BestAvailableEncryption,
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
    load_der_private_key,
    load_pem_private_key,
)

from xorq.common.utils.env_utils import (
    parse_env_file,
)


snowflake_env_var_prefix = "SNOWFLAKE_"
default_env_path = Path(".envrc.secrets.snowflake.keypair")


def make_env_name(name, prefix=snowflake_env_var_prefix):
    return f"{prefix}{name.upper()}"


def make_private_key_pwd(k=20, choices=string.ascii_letters + string.digits):
    return "".join(random.choices(choices, k=k))


def encode_utf8(string):
    return string.encode("utf-8")


def filter_none_one(el):
    # for use with *args
    return filter(None, (el,))


@frozen
class SnowflakeKeypair:
    private_key = field(validator=instance_of(rsa.RSAPrivateKey))
    private_key_pwd = field(validator=instance_of(str), factory=make_private_key_pwd)

    prefix = snowflake_env_var_prefix
    default_path = default_env_path

    def get_private_bytes(
        self, encoding=Encoding.PEM, format=PrivateFormat.PKCS8, encrypted=True
    ):
        return self.private_key.private_bytes(
            encoding=encoding,
            format=format,
            encryption_algorithm=BestAvailableEncryption(
                encode_utf8(self.private_key_pwd)
            )
            if encrypted
            else NoEncryption(),
        )

    @property
    def private_bytes(self):
        return self.get_private_bytes(
            encoding=Encoding.PEM,
            format=PrivateFormat.PKCS8,
            encrypted=True,
        )

    @property
    def private_str(self):
        return self.private_bytes.decode("ascii")

    @property
    def private_str_unencrypted(self):
        return self.get_private_bytes(encrypted=False).decode("ascii")

    @property
    def public_key(self):
        return self.private_key.public_key()

    @property
    def public_bytes(self):
        return self.public_key.public_bytes(
            encoding=Encoding.PEM,
            format=PublicFormat.SubjectPublicKeyInfo,
        )

    @property
    def public_str(self):
        return self.public_bytes.decode("ascii")

    def with_password(self, private_key_pwd):
        return evolve(self, private_key_pwd=private_key_pwd)

    def to_envrc(
        self,
        path=default_path,
        prefix=prefix,
        encrypted=True,
    ):
        names_fields = (
            (
                ("private_key", "private_str"),
                ("public_key", "public_str"),
                ("private_key_pwd", "private_key_pwd"),
            )
            if encrypted
            else (
                ("private_key", "private_str_unencrypted"),
                ("public_key", "public_str"),
            )
        )
        text = "\n".join(
            f"export {name}='{value}'"
            for name, value in (
                (
                    make_env_name(name, prefix=prefix),
                    getattr(self, field),
                )
                for (name, field) in names_fields
            )
        )
        (path := Path(path)).write_text(text)
        return path

    @classmethod
    def generate(cls, password=None):
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )
        return cls(private_key, *filter_none_one(password))

    @classmethod
    def from_bytes_pem(
        cls, private_bytes: bytes, private_key_pwd: Optional[str] = None
    ):
        private_key = load_pem_private_key(
            private_bytes,
            encode_utf8(private_key_pwd) if private_key_pwd else None,
        )
        return cls(private_key, *filter_none_one(private_key_pwd))

    @classmethod
    def from_bytes_der(
        cls, private_bytes: bytes, private_key_pwd: Optional[str] = None
    ):
        private_key = load_der_private_key(
            private_bytes,
            encode_utf8(private_key_pwd) if private_key_pwd else None,
        )
        return cls(private_key, *filter_none_one(private_key_pwd))

    from_bytes = from_bytes_pem

    @classmethod
    def from_str_pem(cls, private_str: str, private_key_pwd: Optional[str] = None):
        return cls.from_bytes(encode_utf8(private_str), private_key_pwd)

    from_str = from_str_pem

    @classmethod
    def from_environment(cls, ctx=os.environ, prefix=prefix):
        kwargs = {
            field: ctx.get(make_env_name(name, prefix=prefix))
            for field, name in (
                ("private_str", "private_key"),
                ("private_key_pwd", "private_key_pwd"),
            )
        }
        return cls.from_str_pem(**kwargs)

    @classmethod
    def from_envrc(cls, path=default_path, prefix=prefix):
        ctx = parse_env_file(path)
        return cls.from_environment(ctx=ctx, prefix=prefix)


def assign_public_key(con, user, public_key_str, do_assert=True):
    from xorq.common.utils.snowflake_utils import execute_statement

    def massage_public_key_str(public_key_str):
        # https://docs.snowflake.com/en/user-guide/key-pair-auth#assign-the-public-key-to-a-snowflake-user
        # # Note: Exclude the public key delimiters in the SQL statement.
        sep = "\n"
        (preamble, *lines, postamble, end) = public_key_str.split(sep)
        assert (preamble, postamble, end) == (
            "-----BEGIN PUBLIC KEY-----",
            "-----END PUBLIC KEY-----",
            "",
        )
        massaged = sep.join(lines)
        return massaged

    massaged_text = massage_public_key_str(public_key_str)
    stmt = f"ALTER USER {user} SET RSA_PUBLIC_KEY='{massaged_text}';"
    fetched = execute_statement(con, stmt, do_assert=do_assert)
    return fetched


def deassign_public_key(con, user, do_assert=True):
    from xorq.common.utils.snowflake_utils import execute_statement

    stmt = f"ALTER USER {user} UNSET RSA_PUBLIC_KEY;"
    fetched = execute_statement(con, stmt, do_assert=do_assert)
    return fetched


def decrypt_private_key_bytes_snowflake(private_key_bytes, password_str):
    # encrypted PEM to unencrypted DER
    private_key = load_pem_private_key(private_key_bytes, password_str.encode("utf-8"))
    return private_key.private_bytes(
        Encoding.DER,
        PrivateFormat.PKCS8,
        NoEncryption(),
    )


def encrypt_private_key_bytes_snowflake_adbc(private_key_bytes, password_str):
    # unencrypted DER to encrypted PEM
    private_key = load_der_private_key(private_key_bytes, None)
    return private_key.private_bytes(
        Encoding.PEM,
        PrivateFormat.PKCS8,
        BestAvailableEncryption(password_str.encode("ascii")),
    )


def maybe_decrypt_private_key(kwargs):
    # snowflake requires DER format for connection
    # ProgrammingError: 251008: Failed to decode private key: Incorrect padding
    # Please provide a valid unencrypted rsa private key in base64-encoded DER format as a str object
    match kwargs:
        case {"private_key": private_key, "private_key_pwd": private_key_pwd, **rest}:
            match private_key:
                case str():
                    private_key = private_key.encode("utf-8")
                case _:
                    raise ValueError("expecting str and only str")
            kwargs = rest | {
                "private_key": decrypt_private_key_bytes_snowflake(
                    private_key, private_key_pwd
                )
            }
        case {"private_key": private_key, **rest}:
            match private_key:
                case str():
                    pass
                case _:
                    warnings.warn(
                        f"expecting str and only str, found: {type(private_key)}",
                        stacklevel=2,
                    )
        case _:
            raise ValueError("private_key must be passed")
    return kwargs


def maybe_encrypt_private_key_snowflake_adbc(private_key, private_key_pwd, N=20):
    match (private_key, private_key_pwd):
        case [str(private_key_encrypted), str()]:
            pass
        case [str(), None]:
            private_key_pwd = "".join(random.choices(string.printable, k=N))
            private_key_encrypted = encrypt_private_key_bytes_snowflake_adbc(
                private_key.encode("utf-8"),
                private_key_pwd,
            )
        case _:
            raise ValueError(
                f"invalid types: {(type(private_key), type(private_key_pwd))}"
            )
    return (private_key_encrypted, private_key_pwd)



from __future__ import annotations

import re

import pandas as pd
import pytest

import xorq.api as xo
import xorq.vendor.ibis.expr.operations as ops
from xorq.backends.conftest import (
    get_cache_uncached,
)
from xorq.backends.snowflake.tests.conftest import (
    generate_mock_get_conn,
    inside_temp_schema,
)
from xorq.caching import (
    ParquetCache,
    SourceSnapshotCache,
)
from xorq.common.utils.snowflake_utils import (
    SnowflakeADBC,
    get_session_query_df,
    get_snowflake_last_modification_time,
)
from xorq.vendor import ibis
from xorq.vendor.ibis.util import gen_name


KEY_PREFIX = xo.config.options.cache.key_prefix


@pytest.mark.snowflake
def test_snowflake_cache_with_name_multiplicity(sf_con):
    (catalog, db) = ("SNOWFLAKE_SAMPLE_DATA", "TPCH_SF1")
    assert sf_con.current_catalog == catalog
    assert sf_con.current_database == db
    table = "CUSTOMER"
    n_tables = (
        sf_con.table("TABLES", database=(catalog, "INFORMATION_SCHEMA"))[
            lambda t: t.TABLE_NAME == table
        ]
        .count()
        .execute()
    )
    assert n_tables > 1
    t = sf_con.table(table)
    (dt,) = t.op().find(ops.DatabaseTable)
    get_snowflake_last_modification_time(dt)


@pytest.mark.snowflake
def test_snowflake_cache_invalidation(sf_con, temp_catalog, temp_db, tmp_path):
    group_by = "key"
    df = pd.DataFrame({group_by: list("abc"), "value": [1, 2, 3]})
    name = gen_name("tmp_table")
    con = xo.connect()
    cache = ParquetCache.from_kwargs(source=con, relative_path=tmp_path)

    # must explicitly invoke USE SCHEMA: use of temp_* DOESN'T impact internal create_table's CREATE TEMP STAGE
    with inside_temp_schema(sf_con, temp_catalog, temp_db):
        table = sf_con.create_table(
            name=name,
            obj=df,
        )
        uncached = table.group_by(group_by).agg(
            {f"min_{col}": table[col].min() for col in table.columns}
        )
        cached_expr = uncached.cache(cache)
        (cache, _) = get_cache_uncached(cached_expr)
        unbound_sql = re.sub(
            r"\s+",
            " ",
            ibis.to_sql(uncached, dialect=sf_con.name),
        )
        query_df = get_session_query_df(sf_con)

        # test preconditions
        assert not cache.exists(uncached)
        assert query_df.QUERY_TEXT.eq(unbound_sql).sum() == 0

        # test cache creation
        xo.execute(cached_expr)
        query_df = get_session_query_df(sf_con)
        assert cache.exists(uncached)
        assert query_df.QUERY_TEXT.eq(unbound_sql).sum() == 1

        # test cache use
        xo.execute(cached_expr)
        assert query_df.QUERY_TEXT.eq(unbound_sql).sum() == 1

        # test cache invalidation
        sf_con.insert(name, df, database=f"{temp_catalog}.{temp_db}")
        assert not cache.exists(uncached)


@pytest.mark.snowflake
def test_snowflake_simple_cache(sf_con, tmp_path):
    db_con = xo.duckdb.connect()
    with inside_temp_schema(sf_con, "SNOWFLAKE_SAMPLE_DATA", "TPCH_SF1"):
        table = sf_con.table("CUSTOMER")
        expr = table.limit(1).cache(
            ParquetCache.from_kwargs(source=db_con, relative_path=tmp_path)
        )
        xo.execute(expr)


@pytest.mark.snowflake
def test_snowflake_native_cache(sf_con, temp_catalog, temp_db, tmp_path):
    group_by = "key"
    df = pd.DataFrame({group_by: list("abc"), "value": [1, 2, 3]})
    name = gen_name("tmp_table")
    cache = ParquetCache.from_kwargs(source=sf_con, relative_path=tmp_path)

    # must explicitly invoke USE SCHEMA: use of temp_* DOESN'T impact internal create_table's CREATE TEMP STAGE
    with inside_temp_schema(sf_con, temp_catalog, temp_db):
        # create a temp table we can mutate
        table = sf_con.create_table(
            name=name,
            obj=df,
        )
        cached_expr = (
            table.group_by(group_by)
            .agg({f"count_{col}": table[col].count() for col in table.columns})
            .cache(cache)
        )
        xo.execute(cached_expr)


@pytest.mark.snowflake
def test_snowflake_snapshot(sf_con, temp_catalog, temp_db):
    group_by = "key"
    df = pd.DataFrame({group_by: list("abc"), "value": [1, 2, 3]})
    name = gen_name("tmp_table")
    cache = SourceSnapshotCache.from_kwargs(source=xo.duckdb.connect())

    # must explicitly invoke USE SCHEMA: use of temp_* DOESN'T impact internal create_table's CREATE TEMP STAGE
    with inside_temp_schema(sf_con, temp_catalog, temp_db):
        # create a temp table we can mutate
        table = sf_con.create_table(
            name=name,
            obj=df,
        )
        uncached = table.group_by(group_by).agg(
            {f"count_{col}": table[col].count() for col in table.columns}
        )
        cached_expr = uncached.cache(cache)
        (cache, _) = get_cache_uncached(cached_expr)
        unbound_sql = re.sub(
            r"\s+",
            " ",
            ibis.to_sql(uncached, dialect=sf_con.name),
        )
        query_df = get_session_query_df(sf_con)

        # test preconditions
        assert not cache.exists(uncached)
        assert query_df.QUERY_TEXT.eq(unbound_sql).sum() == 0

        # test cache creation
        executed0 = xo.execute(cached_expr)
        query_df = get_session_query_df(sf_con)
        assert cache.exists(uncached)
        assert query_df.QUERY_TEXT.eq(unbound_sql).sum() == 1

        # test cache use
        executed1 = xo.execute(cached_expr)
        assert query_df.QUERY_TEXT.eq(unbound_sql).sum() == 1
        assert executed0.equals(executed1)

        # test NO cache invalidation
        sf_con.insert(name, df, database=f"{temp_catalog}.{temp_db}")
        (cache, uncached) = get_cache_uncached(cached_expr)
        assert cache.exists(uncached)
        executed2 = xo.execute(cached_expr.ls.uncached)
        assert not executed0.equals(executed2)


@pytest.mark.snowflake
def test_snowflake_cross_source_native_cache(
    sf_con, pg, temp_catalog, temp_db, tmp_path, mocker
):
    group_by = "number"
    table = pg.table("astronauts")
    cache = ParquetCache.from_kwargs(source=sf_con, relative_path=tmp_path)

    mocker.patch.object(
        SnowflakeADBC,
        "get_conn",
        side_effect=generate_mock_get_conn(
            SnowflakeADBC.get_conn, temp_catalog, temp_db
        ),
        autospec=True,
    )

    with inside_temp_schema(sf_con, temp_catalog, temp_db):
        cached_expr = (
            table.group_by(group_by)
            .agg({f"count_{col}": table[col].count() for col in table.columns})
            .cache(cache)
        )
        actual = cached_expr.execute()

    assert not actual.empty
    assert any(tmp_path.glob(f"{KEY_PREFIX}*")), (
        "The ParquetCache MUST write a parquet file to the given directory"
    )


def test_snowflake_with_failing_name(sf_con, pg, temp_catalog, temp_db, mocker):
    group_by = "tinyint_col"
    table = pg.table("functional_alltypes")

    # caches
    snow_cache = SourceSnapshotCache.from_kwargs(source=sf_con)

    mocker.patch.object(
        SnowflakeADBC,
        "get_conn",
        side_effect=generate_mock_get_conn(
            SnowflakeADBC.get_conn, temp_catalog, temp_db
        ),
        autospec=True,
    )

    with inside_temp_schema(sf_con, temp_catalog, temp_db):
        cached_expr = (
            table.group_by(group_by).agg(xo._.float_col.mean()).cache(snow_cache)
        )
        actual = cached_expr.execute()

    assert not actual.empty



from __future__ import annotations

import datetime
import decimal
import json
import math
import operator


try:
    import regex as re
except ImportError:
    import re
from functools import reduce
from urllib.parse import parse_qs, urlsplit

import numpy as np
import pandas as pd
import toolz

import xorq.vendor.ibis.expr.operations as ops
from xorq.backends.pandas.helpers import isnull


def substring_rowwise(row):
    arg, start, length = row["arg"], row["start"], row["length"]
    if isnull(arg):
        return None
    elif isnull(start):
        return None
    elif isnull(length):
        return arg[start:]
    else:
        return arg[start : start + length]


def substring_serieswise(arg, start, length):
    if length is None:
        return arg.str[start:]
    else:
        return arg.str[start : start + length]


def _sql_like_to_regex(pattern, escape):
    """Convert a SQL `LIKE` pattern to an equivalent Python regular expression.

    Parameters
    ----------
    pattern
        A LIKE pattern with the following semantics:
        * `%` matches zero or more characters
        * `_` matches exactly one character
        * To escape `%` and `_` (or to match the `escape` parameter
          itself), prefix the desired character with `escape`.
    escape
        Escape character

    Returns
    -------
    str
        A regular expression pattern equivalent to the input SQL `LIKE` pattern.

    Examples
    --------
    >>> sql_like_to_regex("6%")  # default is to not escape anything
    '^6.*$'
    >>> sql_like_to_regex("6^%", escape="^")
    '^6%$'
    >>> sql_like_to_regex("6_")
    '^6.$'
    >>> sql_like_to_regex("6/_", escape="/")
    '^6_$'
    >>> sql_like_to_regex("%abc")  # any string ending with "abc"
    '^.*abc$'
    >>> sql_like_to_regex("abc%")  # any string starting with "abc"
    '^abc.*$'

    """
    cur_i = 0
    pattern_length = len(pattern)

    while cur_i < pattern_length:
        nxt_i = cur_i + 1

        cur = pattern[cur_i]
        nxt = pattern[nxt_i] if nxt_i < pattern_length else None

        skip = 1

        if nxt is not None and escape is not None and cur == escape:
            yield nxt
            skip = 2
        elif cur == "%":
            yield ".*"
        elif cur == "_":
            yield "."
        else:
            yield cur

        cur_i += skip


def sql_like_to_regex(pattern, escape=None):
    return f"^{''.join(_sql_like_to_regex(pattern, escape))}$"


def string_sqllike_serieswise(arg, pattern, escape):
    pat = sql_like_to_regex(pattern, escape)
    return arg.str.contains(pat, regex=True)


def string_sqlilike_serieswise(arg, pattern, escape):
    pat = sql_like_to_regex(pattern, escape)
    return arg.str.contains(pat, regex=True, flags=re.IGNORECASE)


def extract_userinfo_elementwise(x):
    url_parts = urlsplit(x)
    username = url_parts.username or ""
    password = url_parts.password or ""
    return f"{username}:{password}"


def extract_queryparam_rowwise(row):
    query = urlsplit(row["arg"]).query
    param_name = row["key"]
    if param_name is not None:
        value = parse_qs(query)[param_name]
        return value if len(value) > 1 else value[0]
    else:
        return query


def array_index_rowwise(row):
    try:
        return row["arg"][row["index"]]
    except IndexError:
        return None


def array_position_rowwise(row):
    try:
        return row["arg"].index(row["other"])
    except ValueError:
        return -1


def array_remove_rowwise(row):
    if row["arg"] is None:
        return None
    return [x for x in row["arg"] if x != row["other"]]


def array_slice_rowwise(row):
    arg, start, stop = row["arg"], row["start"], row["stop"]
    if isnull(start) and isnull(stop):
        return arg
    elif isnull(start):
        return arg[:stop]
    elif isnull(stop):
        return arg[start:]
    else:
        return arg[start:stop]


def integer_range_rowwise(row):
    if not row["step"]:
        return []
    return list(np.arange(row["start"], row["stop"], row["step"]))


def timestamp_range_rowwise(row):
    if not row["step"]:
        return []
    return list(
        pd.date_range(row["start"], row["stop"], freq=row["step"], inclusive="left")
    )


def _safe_method(mapping, method, *args, **kwargs):
    if isnull(mapping):
        return None
    try:
        method = getattr(mapping, method)
    except AttributeError:
        return None
    else:
        result = method(*args, **kwargs)
        return None if isnull(result) else result


def safe_len(mapping):
    return _safe_method(mapping, "__len__")


def safe_get(mapping, key, default=None):
    return _safe_method(mapping, "get", key, default)


def safe_contains(mapping, key):
    return _safe_method(mapping, "__contains__", key)


def safe_keys(mapping):
    result = _safe_method(mapping, "keys")
    if result is None:
        return None
    # list(...) to unpack iterable
    return np.array(list(result))


def safe_values(mapping):
    result = _safe_method(mapping, "values")
    if result is None or result is pd.NA:
        return None
    # list(...) to unpack iterable
    return np.array(list(result), dtype="object")


def safe_merge(left, right):
    if isnull(left) or isnull(right):
        return None
    else:
        return {**left, **right}


def safe_json_getitem(value, key):
    try:
        # try to deserialize the value -> return None if it's None
        if (js := json.loads(value)) is None:
            return None
    except (json.JSONDecodeError, TypeError):
        # if there's an error related to decoding or a type error return None
        return None

    try:
        # try to extract the value as an array element or mapping key
        return js[key]
    except (KeyError, IndexError, TypeError):
        # KeyError: missing mapping key
        # IndexError: missing sequence key
        # TypeError: `js` doesn't implement __getitem__, either at all or for
        # the type of `key`
        return None


def safe_decimal(func):
    def wrapper(x, **kwargs):
        try:
            return func(x, **kwargs)
        except decimal.InvalidOperation:
            return decimal.Decimal("NaN")

    return wrapper


def round_serieswise(arg, digits):
    if digits is None:
        return np.round(arg).astype("int64")
    else:
        return np.round(arg, digits).astype("float64")


def arbitrary(arg):
    arg = arg.dropna()
    return arg.iat[0] if len(arg) else None


reductions = {
    ops.Min: lambda x: x.min(),
    ops.Max: lambda x: x.max(),
    ops.Sum: lambda x: x.sum(),
    ops.Mean: lambda x: x.mean(),
    ops.Count: lambda x: x.count(),
    ops.CountStar: lambda x: x.count(),
    ops.Mode: lambda x: x.mode().iat[0],
    ops.Any: lambda x: x.any(),
    ops.All: lambda x: x.all(),
    ops.Median: lambda x: x.median(),
    ops.ApproxMedian: lambda x: x.median(),
    ops.BitAnd: lambda x: np.bitwise_and.reduce(x.values),
    ops.BitOr: lambda x: np.bitwise_or.reduce(x.values),
    ops.BitXor: lambda x: np.bitwise_xor.reduce(x.values),
    ops.Arbitrary: arbitrary,
    ops.CountDistinct: lambda x: x.nunique(),
    ops.ApproxCountDistinct: lambda x: x.nunique(),
}


_generic = {
    ops.Abs: abs,
    ops.Acos: np.arccos,
    ops.Add: operator.add,
    ops.And: operator.and_,
    ops.Asin: np.arcsin,
    ops.Atan: np.arctan,
    ops.Atan2: np.arctan2,
    ops.BitwiseAnd: lambda x, y: np.bitwise_and(x, y),
    ops.BitwiseLeftShift: lambda x, y: np.left_shift(x, y).astype("int64"),
    ops.BitwiseNot: np.invert,
    ops.BitwiseOr: lambda x, y: np.bitwise_or(x, y),
    ops.BitwiseRightShift: lambda x, y: np.right_shift(x, y).astype("int64"),
    ops.BitwiseXor: lambda x, y: np.bitwise_xor(x, y),
    ops.Ceil: lambda x: np.ceil(x).astype("int64"),
    ops.Cos: np.cos,
    ops.Cot: lambda x: 1 / np.tan(x),
    ops.DateAdd: operator.add,
    ops.DateDiff: operator.sub,
    ops.DateSub: operator.sub,
    ops.Degrees: np.degrees,
    ops.Divide: operator.truediv,
    ops.Equals: operator.eq,
    ops.Exp: np.exp,
    ops.Floor: lambda x: np.floor(x).astype("int64"),
    ops.FloorDivide: operator.floordiv,
    ops.Greater: operator.gt,
    ops.GreaterEqual: operator.ge,
    ops.IdenticalTo: lambda x, y: (x == y) | (pd.isnull(x) & pd.isnull(y)),
    ops.IntervalAdd: operator.add,
    ops.IntervalFloorDivide: operator.floordiv,
    ops.IntervalMultiply: operator.mul,
    ops.IntervalSubtract: operator.sub,
    ops.Less: operator.lt,
    ops.LessEqual: operator.le,
    ops.Ln: np.log,
    ops.Log10: np.log10,
    ops.Log2: np.log2,
    ops.Modulus: operator.mod,
    ops.Multiply: operator.mul,
    ops.Negate: lambda x: not x if isinstance(x, (bool, np.bool_)) else -x,
    ops.Not: lambda x: not x if isinstance(x, (bool, np.bool_)) else ~x,
    ops.NotEquals: operator.ne,
    ops.Or: operator.or_,
    ops.Power: operator.pow,
    ops.Radians: np.radians,
    ops.Sign: np.sign,
    ops.Sin: np.sin,
    ops.Sqrt: np.sqrt,
    ops.Subtract: operator.sub,
    ops.Tan: np.tan,
    ops.TimestampAdd: operator.add,
    ops.TimestampDiff: operator.sub,
    ops.TimestampSub: operator.sub,
    ops.Xor: operator.xor,
    ops.E: lambda: np.e,
    ops.Pi: lambda: np.pi,
    ops.TimestampNow: lambda: pd.Timestamp("now", tz="UTC").tz_localize(None),
    ops.DateNow: lambda: pd.Timestamp(datetime.date.today()),
    ops.StringConcat: lambda xs: reduce(operator.add, xs),
    ops.StringJoin: lambda xs, sep: reduce(lambda x, y: x + sep + y, xs),
    ops.Log: lambda x, base: np.log(x) if base is None else np.log(x) / np.log(base),
}


def none_safe(func):
    def wrapper(*args, **kwargs):
        if any(map(isnull, args)):
            return None
        return func(*args, **kwargs)

    return wrapper


generic = {
    **{k: none_safe(v) for k, v in _generic.items()},
    ops.IsNull: pd.isnull,
    ops.NotNull: pd.notnull,
    ops.IsInf: np.isinf,
}


columnwise = {
    ops.Clip: lambda df: df["arg"].clip(lower=df["lower"], upper=df["upper"]),
    ops.IfElse: lambda df: df["true_expr"].where(
        df["bool_expr"], other=df["false_null_expr"]
    ),
    ops.NullIf: lambda df: df["arg"].where(df["arg"] != df["null_if_expr"]),
    ops.Repeat: lambda df: df["arg"] * df["times"],
}


rowwise = {
    ops.ArrayContains: lambda row: row["other"] in row["arg"],
    ops.ArrayIndex: array_index_rowwise,
    ops.ArrayPosition: array_position_rowwise,
    ops.ArrayRemove: array_remove_rowwise,
    ops.ArrayRepeat: lambda row: np.tile(row["arg"], max(0, row["times"])),
    ops.ArraySlice: array_slice_rowwise,
    ops.ArrayUnion: lambda row: toolz.unique(row["left"] + row["right"]),
    ops.EndsWith: lambda row: row["arg"].endswith(row["end"]),
    ops.IntegerRange: integer_range_rowwise,
    ops.JSONGetItem: lambda row: safe_json_getitem(row["arg"], row["index"]),
    ops.Map: lambda row: dict(zip(row["keys"], row["values"])),
    ops.MapGet: lambda row: safe_get(row["arg"], row["key"], row["default"]),
    ops.MapContains: lambda row: safe_contains(row["arg"], row["key"]),
    ops.MapMerge: lambda row: safe_merge(row["left"], row["right"]),
    ops.TimestampRange: timestamp_range_rowwise,
    ops.LPad: lambda row: row["arg"].rjust(row["length"], row["pad"]),
    ops.RegexExtract: lambda row: re.search(row["pattern"], row["arg"]).group(
        row["index"]
    ),
    ops.RegexReplace: lambda row: re.sub(
        row["pattern"], row["replacement"], row["arg"]
    ),
    ops.RegexSearch: lambda row: re.search(row["pattern"], row["arg"]) is not None,
    ops.RPad: lambda row: row["arg"].ljust(row["length"], row["pad"]),
    ops.StartsWith: lambda row: row["arg"].startswith(row["start"]),
    ops.StringContains: lambda row: row["haystack"].contains(row["needle"]),
    ops.StringFind: lambda row: row["arg"].find(
        row["substr"], row["start"], row["end"]
    ),
    ops.StringReplace: lambda row: row["arg"].replace(
        row["pattern"], row["replacement"]
    ),
    ops.StringSplit: lambda row: row["arg"].split(row["delimiter"]),
    ops.StrRight: lambda row: row["arg"][-row["nchars"] :],
    ops.Translate: lambda row: row["arg"].translate(
        str.maketrans(row["from_str"], row["to_str"])
    ),
    ops.Substring: substring_rowwise,
    ops.ExtractQuery: extract_queryparam_rowwise,
    ops.Strftime: lambda row: row["arg"].strftime(row["format_str"]),
}


def custom_hash(arg):
    import pyarrow as pa

    from xorq.internal import SessionContext

    name = "custom_hash"
    table_name = f"table_{name}"
    row_id_name = f"row_id_{name}"

    ctx = SessionContext()
    ctx.register_record_batches(
        table_name,
        [
            pa.Table.from_pandas(
                arg.rename(name).to_frame().reset_index(names=row_id_name)
            ).to_batches()
        ],
    )

    return (
        ctx.sql(
            f"select hash_int({name}) as col_name from {table_name} order by {row_id_name}"
        )
        .to_pandas()
        .squeeze()
    )


serieswise = {
    ops.Between: lambda arg, lower_bound, upper_bound: arg.between(
        lower_bound, upper_bound
    ),
    ops.Capitalize: lambda arg: arg.str.capitalize(),
    ops.Date: lambda arg: arg.dt.floor("d"),
    ops.DayOfWeekIndex: lambda arg: pd.to_datetime(arg).dt.dayofweek,
    ops.DayOfWeekName: lambda arg: pd.to_datetime(arg).dt.day_name(),
    ops.EndsWith: lambda arg, end: arg.str.endswith(end),
    ops.ExtractDay: lambda arg: arg.dt.day,
    ops.ExtractDayOfYear: lambda arg: arg.dt.dayofyear,
    ops.ExtractEpochSeconds: lambda arg: (
        arg.astype("datetime64[s]").astype("int64").astype("int32")
    ),
    ops.ExtractHour: lambda arg: arg.dt.hour,
    ops.ExtractMicrosecond: lambda arg: arg.dt.microsecond,
    ops.ExtractMillisecond: lambda arg: arg.dt.microsecond // 1000,
    ops.ExtractMinute: lambda arg: arg.dt.minute,
    ops.ExtractMonth: lambda arg: arg.dt.month,
    ops.ExtractQuarter: lambda arg: arg.dt.quarter,
    ops.ExtractSecond: lambda arg: arg.dt.second,
    ops.ExtractWeekOfYear: lambda arg: arg.dt.isocalendar().week.astype("int32"),
    ops.ExtractYear: lambda arg: arg.dt.year,
    ops.ExtractIsoYear: lambda arg: arg.dt.isocalendar().year,
    ops.IsNull: lambda arg: arg.isnull(),
    ops.NotNull: lambda arg: arg.notnull(),
    ops.Lowercase: lambda arg: arg.str.lower(),
    ops.LPad: lambda arg, length, pad: arg.str.rjust(length, fillchar=pad),
    ops.LStrip: lambda arg: arg.str.lstrip(),
    ops.Repeat: lambda arg, times: arg.str.repeat(times),
    ops.Reverse: lambda arg: arg.str[::-1],
    ops.Round: round_serieswise,
    ops.RPad: lambda arg, length, pad: arg.str.ljust(length, fillchar=pad),
    ops.RStrip: lambda arg: arg.str.rstrip(),
    ops.StartsWith: lambda arg, start: arg.str.startswith(start),
    ops.StringAscii: lambda arg: arg.map(ord, na_action="ignore").astype("int32"),
    ops.StringContains: lambda haystack, needle: haystack.str.contains(
        needle, regex=False
    ),
    ops.StringFind: lambda arg, substr, start, end: arg.str.find(substr, start, end),
    ops.StringLength: lambda arg: arg.str.len().astype("int32"),
    ops.StringReplace: lambda arg, pattern, replacement: arg.str.replace(
        pattern, replacement
    ),
    ops.StringSplit: lambda arg, delimiter: arg.str.split(delimiter),
    ops.StringSQLLike: string_sqllike_serieswise,
    ops.StringSQLILike: string_sqlilike_serieswise,
    ops.Strip: lambda arg: arg.str.strip(),
    ops.Strftime: lambda arg, format_str: arg.dt.strftime(format_str),
    ops.StrRight: lambda arg, nchars: arg.str[-nchars:],
    ops.Substring: substring_serieswise,
    ops.Time: lambda arg: arg.dt.time,
    ops.TimestampFromUNIX: lambda arg, unit: pd.to_datetime(arg, unit=unit.short),
    ops.Translate: lambda arg, from_str, to_str: arg.str.translate(
        str.maketrans(from_str, to_str)
    ),
    ops.Uppercase: lambda arg: arg.str.upper(),
    ops.Hash: custom_hash,
}

elementwise = {
    ops.ExtractProtocol: lambda x: getattr(urlsplit(x), "scheme", ""),
    ops.ExtractAuthority: lambda x: getattr(urlsplit(x), "netloc", ""),
    ops.ExtractPath: lambda x: getattr(urlsplit(x), "path", ""),
    ops.ExtractFragment: lambda x: getattr(urlsplit(x), "fragment", ""),
    ops.ExtractHost: lambda x: getattr(urlsplit(x), "hostname", ""),
    ops.ExtractUserInfo: extract_userinfo_elementwise,
    ops.StructField: lambda x, field: safe_get(x, field),
    ops.ArrayLength: len,
    ops.ArrayFlatten: toolz.concat,
    ops.ArraySort: sorted,
    ops.ArrayDistinct: toolz.unique,
    ops.MapLength: safe_len,
    ops.MapKeys: safe_keys,
    ops.MapValues: safe_values,
    ops.Round: lambda x, digits=0: round(x, digits),
}


elementwise_decimal = {
    ops.Round: lambda x, digits=0: round(x, digits),
    ops.Log10: safe_decimal(lambda x: x.log10()),
    ops.Ln: safe_decimal(lambda x: x.ln()),
    ops.Exp: safe_decimal(lambda x: x.exp()),
    ops.Floor: safe_decimal(math.floor),
    ops.Ceil: safe_decimal(math.ceil),
    ops.Sqrt: safe_decimal(lambda x: x.sqrt()),
    ops.Log2: safe_decimal(lambda x: x.ln() / decimal.Decimal(2).ln()),
    ops.Sign: safe_decimal(lambda x: math.copysign(1, x)),
    ops.Log: safe_decimal(lambda x, base: x.ln() / decimal.Decimal(base).ln()),
    ops.Multiply: safe_decimal(lambda left, right: left * right),
}


supported_operations = (
    generic.keys()
    | columnwise.keys()
    | rowwise.keys()
    | serieswise.keys()
    | elementwise.keys()
)
